# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import decimal
import zaber_bson


@dataclass
class LockstepAxes:
    """
    The axis numbers of a lockstep group.
    """

    axis_1: int
    """
    The axis number used to set the first axis.
    """

    axis_2: int
    """
    The axis number used to set the second axis.
    """

    axis_3: int
    """
    The axis number used to set the third axis.
    """

    axis_4: int
    """
    The axis number used to set the fourth axis.
    """

    @staticmethod
    def zero_values() -> 'LockstepAxes':
        return LockstepAxes(
            axis_1=0,
            axis_2=0,
            axis_3=0,
            axis_4=0,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'LockstepAxes':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return LockstepAxes.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'axis1': int(self.axis_1),
            'axis2': int(self.axis_2),
            'axis3': int(self.axis_3),
            'axis4': int(self.axis_4),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'LockstepAxes':
        return LockstepAxes(
            axis_1=data.get('axis1'),  # type: ignore
            axis_2=data.get('axis2'),  # type: ignore
            axis_3=data.get('axis3'),  # type: ignore
            axis_4=data.get('axis4'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.axis_1 is None:
            raise ValueError(f'Property "Axis1" of "LockstepAxes" is None.')

        if not isinstance(self.axis_1, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Axis1" of "LockstepAxes" is not a number.')

        if int(self.axis_1) != self.axis_1:
            raise ValueError(f'Property "Axis1" of "LockstepAxes" is not integer value.')

        if self.axis_2 is None:
            raise ValueError(f'Property "Axis2" of "LockstepAxes" is None.')

        if not isinstance(self.axis_2, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Axis2" of "LockstepAxes" is not a number.')

        if int(self.axis_2) != self.axis_2:
            raise ValueError(f'Property "Axis2" of "LockstepAxes" is not integer value.')

        if self.axis_3 is None:
            raise ValueError(f'Property "Axis3" of "LockstepAxes" is None.')

        if not isinstance(self.axis_3, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Axis3" of "LockstepAxes" is not a number.')

        if int(self.axis_3) != self.axis_3:
            raise ValueError(f'Property "Axis3" of "LockstepAxes" is not integer value.')

        if self.axis_4 is None:
            raise ValueError(f'Property "Axis4" of "LockstepAxes" is None.')

        if not isinstance(self.axis_4, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Axis4" of "LockstepAxes" is not a number.')

        if int(self.axis_4) != self.axis_4:
            raise ValueError(f'Property "Axis4" of "LockstepAxes" is not integer value.')



# This file is generated from the Zaber device database. Do not manually edit this file.
# pylint: disable=line-too-long, too-many-branches, too-many-return-statements

from enum import Enum
from typing import Union, Literal


class Units(Enum):
    NATIVE = ''

    LENGTH_METRES = 'Length:metres'
    LENGTH_CENTIMETRES = 'Length:centimetres'
    LENGTH_MILLIMETRES = 'Length:millimetres'
    LENGTH_MICROMETRES = 'Length:micrometres'
    LENGTH_NANOMETRES = 'Length:nanometres'
    LENGTH_INCHES = 'Length:inches'

    VELOCITY_METRES_PER_SECOND = 'Velocity:metres per second'
    VELOCITY_CENTIMETRES_PER_SECOND = 'Velocity:centimetres per second'
    VELOCITY_MILLIMETRES_PER_SECOND = 'Velocity:millimetres per second'
    VELOCITY_MICROMETRES_PER_SECOND = 'Velocity:micrometres per second'
    VELOCITY_NANOMETRES_PER_SECOND = 'Velocity:nanometres per second'
    VELOCITY_INCHES_PER_SECOND = 'Velocity:inches per second'

    ACCELERATION_METRES_PER_SECOND_SQUARED = 'Acceleration:metres per second squared'
    ACCELERATION_CENTIMETRES_PER_SECOND_SQUARED = 'Acceleration:centimetres per second squared'
    ACCELERATION_MILLIMETRES_PER_SECOND_SQUARED = 'Acceleration:millimetres per second squared'
    ACCELERATION_MICROMETRES_PER_SECOND_SQUARED = 'Acceleration:micrometres per second squared'
    ACCELERATION_NANOMETRES_PER_SECOND_SQUARED = 'Acceleration:nanometres per second squared'
    ACCELERATION_INCHES_PER_SECOND_SQUARED = 'Acceleration:inches per second squared'

    ANGLE_DEGREES = 'Angle:degrees'
    ANGLE_RADIANS = 'Angle:radians'

    ANGULAR_VELOCITY_DEGREES_PER_SECOND = 'Angular Velocity:degrees per second'
    ANGULAR_VELOCITY_RADIANS_PER_SECOND = 'Angular Velocity:radians per second'

    ANGULAR_ACCELERATION_DEGREES_PER_SECOND_SQUARED = 'Angular Acceleration:degrees per second squared'
    ANGULAR_ACCELERATION_RADIANS_PER_SECOND_SQUARED = 'Angular Acceleration:radians per second squared'

    AC_ELECTRIC_CURRENT_AMPERES_PEAK = 'AC Electric Current:amperes peak'
    AC_ELECTRIC_CURRENT_AMPERES_RMS = 'AC Electric Current:amperes RMS'

    PERCENT_PERCENT = 'Percent:percent'

    DC_ELECTRIC_CURRENT_AMPERES = 'DC Electric Current:amperes'

    FORCE_NEWTONS = 'Force:newtons'
    FORCE_MILLINEWTONS = 'Force:millinewtons'
    FORCE_POUNDS_FORCE = 'Force:pounds-force'
    FORCE_KILONEWTONS = 'Force:kilonewtons'

    TIME_SECONDS = 'Time:seconds'
    TIME_MILLISECONDS = 'Time:milliseconds'
    TIME_MICROSECONDS = 'Time:microseconds'

    TORQUE_NEWTON_METRES = 'Torque:newton metres'
    TORQUE_NEWTON_CENTIMETRES = 'Torque:newton centimetres'
    TORQUE_POUND_FORCE_FEET = 'Torque:pound-force-feet'
    TORQUE_OUNCE_FORCE_INCHES = 'Torque:ounce-force-inches'

    INERTIA_GRAMS = 'Inertia:grams'
    INERTIA_KILOGRAMS = 'Inertia:kilograms'
    INERTIA_MILLIGRAMS = 'Inertia:milligrams'
    INERTIA_POUNDS = 'Inertia:pounds'
    INERTIA_OUNCES = 'Inertia:ounces'

    ROTATIONAL_INERTIA_GRAM_SQUARE_METRE = 'Rotational Inertia:gram-square metre'
    ROTATIONAL_INERTIA_KILOGRAM_SQUARE_METRE = 'Rotational Inertia:kilogram-square metre'
    ROTATIONAL_INERTIA_POUND_SQUARE_FEET = 'Rotational Inertia:pound-square-feet'

    FORCE_CONSTANT_NEWTONS_PER_AMP = 'Force Constant:newtons per amp'
    FORCE_CONSTANT_MILLINEWTONS_PER_AMP = 'Force Constant:millinewtons per amp'
    FORCE_CONSTANT_KILONEWTONS_PER_AMP = 'Force Constant:kilonewtons per amp'
    FORCE_CONSTANT_POUNDS_FORCE_PER_AMP = 'Force Constant:pounds-force per amp'

    TORQUE_CONSTANT_NEWTON_METRES_PER_AMP = 'Torque Constant:newton metres per amp'
    TORQUE_CONSTANT_MILLINEWTON_METRES_PER_AMP = 'Torque Constant:millinewton metres per amp'
    TORQUE_CONSTANT_KILONEWTON_METRES_PER_AMP = 'Torque Constant:kilonewton metres per amp'
    TORQUE_CONSTANT_POUND_FORCE_FEET_PER_AMP = 'Torque Constant:pound-force-feet per amp'

    VOLTAGE_VOLTS = 'Voltage:volts'
    VOLTAGE_MILLIVOLTS = 'Voltage:millivolts'
    VOLTAGE_MICROVOLTS = 'Voltage:microvolts'

    CURRENT_CONTROLLER_PROPORTIONAL_GAIN_VOLTS_PER_AMP = 'Current Controller Proportional Gain:volts per amp'
    CURRENT_CONTROLLER_PROPORTIONAL_GAIN_MILLIVOLTS_PER_AMP = 'Current Controller Proportional Gain:millivolts per amp'
    CURRENT_CONTROLLER_PROPORTIONAL_GAIN_MICROVOLTS_PER_AMP = 'Current Controller Proportional Gain:microvolts per amp'

    CURRENT_CONTROLLER_INTEGRAL_GAIN_VOLTS_PER_AMP_PER_SECOND = 'Current Controller Integral Gain:volts per amp per second'
    CURRENT_CONTROLLER_INTEGRAL_GAIN_MILLIVOLTS_PER_AMP_PER_SECOND = 'Current Controller Integral Gain:millivolts per amp per second'
    CURRENT_CONTROLLER_INTEGRAL_GAIN_MICROVOLTS_PER_AMP_PER_SECOND = 'Current Controller Integral Gain:microvolts per amp per second'

    CURRENT_CONTROLLER_DERIVATIVE_GAIN_VOLTS_SECOND_PER_AMP = 'Current Controller Derivative Gain:volts second per amp'
    CURRENT_CONTROLLER_DERIVATIVE_GAIN_MILLIVOLTS_SECOND_PER_AMP = 'Current Controller Derivative Gain:millivolts second per amp'
    CURRENT_CONTROLLER_DERIVATIVE_GAIN_MICROVOLTS_SECOND_PER_AMP = 'Current Controller Derivative Gain:microvolts second per amp'

    RESISTANCE_KILOOHMS = 'Resistance:kiloohms'
    RESISTANCE_OHMS = 'Resistance:ohms'
    RESISTANCE_MILLIOHMS = 'Resistance:milliohms'
    RESISTANCE_MICROOHMS = 'Resistance:microohms'
    RESISTANCE_NANOOHMS = 'Resistance:nanoohms'

    INDUCTANCE_HENRIES = 'Inductance:henries'
    INDUCTANCE_MILLIHENRIES = 'Inductance:millihenries'
    INDUCTANCE_MICROHENRIES = 'Inductance:microhenries'
    INDUCTANCE_NANOHENRIES = 'Inductance:nanohenries'

    VOLTAGE_CONSTANT_VOLT_SECONDS_PER_RADIAN = 'Voltage Constant:volt seconds per radian'
    VOLTAGE_CONSTANT_MILLIVOLT_SECONDS_PER_RADIAN = 'Voltage Constant:millivolt seconds per radian'
    VOLTAGE_CONSTANT_MICROVOLT_SECONDS_PER_RADIAN = 'Voltage Constant:microvolt seconds per radian'

    ABSOLUTE_TEMPERATURE_DEGREES_CELSIUS = 'Absolute Temperature:degrees Celsius'
    ABSOLUTE_TEMPERATURE_KELVINS = 'Absolute Temperature:kelvins'
    ABSOLUTE_TEMPERATURE_DEGREES_FAHRENHEIT = 'Absolute Temperature:degrees Fahrenheit'
    ABSOLUTE_TEMPERATURE_DEGREES_RANKINE = 'Absolute Temperature:degrees Rankine'

    RELATIVE_TEMPERATURE_DEGREES_CELSIUS = 'Relative Temperature:degrees Celsius'
    RELATIVE_TEMPERATURE_KELVINS = 'Relative Temperature:kelvins'
    RELATIVE_TEMPERATURE_DEGREES_FAHRENHEIT = 'Relative Temperature:degrees Fahrenheit'
    RELATIVE_TEMPERATURE_DEGREES_RANKINE = 'Relative Temperature:degrees Rankine'

    FREQUENCY_GIGAHERTZ = 'Frequency:gigahertz'
    FREQUENCY_MEGAHERTZ = 'Frequency:megahertz'
    FREQUENCY_KILOHERTZ = 'Frequency:kilohertz'
    FREQUENCY_HERTZ = 'Frequency:hertz'
    FREQUENCY_MILLIHERTZ = 'Frequency:millihertz'
    FREQUENCY_MICROHERTZ = 'Frequency:microhertz'
    FREQUENCY_NANOHERTZ = 'Frequency:nanohertz'


UnitsAndLiterals = Union[Units, Literal["m", "cm", "mm", "µm", "um", "nm", "in", "m/s", "cm/s", "mm/s", "µm/s", "um/s", "nm/s", "in/s", "m/s²", "m/s^2", "cm/s²", "cm/s^2", "mm/s²", "mm/s^2", "µm/s²", "um/s^2", "nm/s²", "nm/s^2", "in/s²", "in/s^2", "°", "deg", "rad", "°/s", "deg/s", "rad/s", "°/s²", "deg/s^2", "rad/s²", "rad/s^2", "%", "s", "ms", "µs", "us", "GHz", "MHz", "kHz", "Hz", "mHz", "µHz", "uHz", "nHz"]]
LengthUnits = Union[Units, Literal["m", "cm", "mm", "µm", "um", "nm", "in", "°", "deg", "rad"]]
VelocityUnits = Union[Units, Literal["m/s", "cm/s", "mm/s", "µm/s", "um/s", "nm/s", "in/s", "°/s", "deg/s", "rad/s"]]
AccelerationUnits = Union[Units, Literal["m/s²", "m/s^2", "cm/s²", "cm/s^2", "mm/s²", "mm/s^2", "µm/s²", "um/s^2", "nm/s²", "nm/s^2", "in/s²", "in/s^2", "°/s²", "deg/s^2", "rad/s²", "rad/s^2"]]
TimeUnits = Union[Units, Literal["s", "ms", "µs", "us"]]
FrequencyUnits = Union[Units, Literal["GHz", "MHz", "kHz", "Hz", "mHz", "µHz", "uHz", "nHz"]]

LITERALS_TO_UNITS = {
    "m": Units.LENGTH_METRES,
    "cm": Units.LENGTH_CENTIMETRES,
    "mm": Units.LENGTH_MILLIMETRES,
    "µm": Units.LENGTH_MICROMETRES,
    "um": Units.LENGTH_MICROMETRES,
    "nm": Units.LENGTH_NANOMETRES,
    "in": Units.LENGTH_INCHES,
    "m/s": Units.VELOCITY_METRES_PER_SECOND,
    "cm/s": Units.VELOCITY_CENTIMETRES_PER_SECOND,
    "mm/s": Units.VELOCITY_MILLIMETRES_PER_SECOND,
    "µm/s": Units.VELOCITY_MICROMETRES_PER_SECOND,
    "um/s": Units.VELOCITY_MICROMETRES_PER_SECOND,
    "nm/s": Units.VELOCITY_NANOMETRES_PER_SECOND,
    "in/s": Units.VELOCITY_INCHES_PER_SECOND,
    "m/s²": Units.ACCELERATION_METRES_PER_SECOND_SQUARED,
    "m/s^2": Units.ACCELERATION_METRES_PER_SECOND_SQUARED,
    "cm/s²": Units.ACCELERATION_CENTIMETRES_PER_SECOND_SQUARED,
    "cm/s^2": Units.ACCELERATION_CENTIMETRES_PER_SECOND_SQUARED,
    "mm/s²": Units.ACCELERATION_MILLIMETRES_PER_SECOND_SQUARED,
    "mm/s^2": Units.ACCELERATION_MILLIMETRES_PER_SECOND_SQUARED,
    "µm/s²": Units.ACCELERATION_MICROMETRES_PER_SECOND_SQUARED,
    "um/s^2": Units.ACCELERATION_MICROMETRES_PER_SECOND_SQUARED,
    "nm/s²": Units.ACCELERATION_NANOMETRES_PER_SECOND_SQUARED,
    "nm/s^2": Units.ACCELERATION_NANOMETRES_PER_SECOND_SQUARED,
    "in/s²": Units.ACCELERATION_INCHES_PER_SECOND_SQUARED,
    "in/s^2": Units.ACCELERATION_INCHES_PER_SECOND_SQUARED,
    "°": Units.ANGLE_DEGREES,
    "deg": Units.ANGLE_DEGREES,
    "rad": Units.ANGLE_RADIANS,
    "°/s": Units.ANGULAR_VELOCITY_DEGREES_PER_SECOND,
    "deg/s": Units.ANGULAR_VELOCITY_DEGREES_PER_SECOND,
    "rad/s": Units.ANGULAR_VELOCITY_RADIANS_PER_SECOND,
    "°/s²": Units.ANGULAR_ACCELERATION_DEGREES_PER_SECOND_SQUARED,
    "deg/s^2": Units.ANGULAR_ACCELERATION_DEGREES_PER_SECOND_SQUARED,
    "rad/s²": Units.ANGULAR_ACCELERATION_RADIANS_PER_SECOND_SQUARED,
    "rad/s^2": Units.ANGULAR_ACCELERATION_RADIANS_PER_SECOND_SQUARED,
    "%": Units.PERCENT_PERCENT,
    "s": Units.TIME_SECONDS,
    "ms": Units.TIME_MILLISECONDS,
    "µs": Units.TIME_MICROSECONDS,
    "us": Units.TIME_MICROSECONDS,
    "GHz": Units.FREQUENCY_GIGAHERTZ,
    "MHz": Units.FREQUENCY_MEGAHERTZ,
    "kHz": Units.FREQUENCY_KILOHERTZ,
    "Hz": Units.FREQUENCY_HERTZ,
    "mHz": Units.FREQUENCY_MILLIHERTZ,
    "µHz": Units.FREQUENCY_MICROHERTZ,
    "uHz": Units.FREQUENCY_MICROHERTZ,
    "nHz": Units.FREQUENCY_NANOHERTZ,
}


def units_from_literals(units: UnitsAndLiterals) -> Units:
    if isinstance(units, Units):
        return units

    converted = LITERALS_TO_UNITS.get(units)
    if converted is None:
        raise ValueError(f"Invalid units: {units}")

    return converted



# pylint: disable=too-many-arguments, too-many-lines

# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #
from typing import TYPE_CHECKING, List
from ..dto import requests as dto
from ..units import Units, TimeUnits
from ..call import call, call_async
from ..dto.ascii.digital_output_action import DigitalOutputAction

if TYPE_CHECKING:
    from .device import Device


class StreamIo:
    """
    Class providing access to I/O for a stream.
    """

    def __init__(self, device: 'Device', stream_id: int):
        self._device = device
        self._stream_id = stream_id

    def set_digital_output(
            self,
            channel_number: int,
            value: DigitalOutputAction
    ) -> None:
        """
        Sets value for the specified digital output channel.

        Args:
            channel_number: Channel number starting at 1.
            value: The type of action to perform on the channel.
        """
        request = dto.StreamSetDigitalOutputRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            stream_id=self._stream_id,
            channel_number=channel_number,
            value=value,
        )
        call("device/stream_set_digital_output", request)

    async def set_digital_output_async(
            self,
            channel_number: int,
            value: DigitalOutputAction
    ) -> None:
        """
        Sets value for the specified digital output channel.

        Args:
            channel_number: Channel number starting at 1.
            value: The type of action to perform on the channel.
        """
        request = dto.StreamSetDigitalOutputRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            stream_id=self._stream_id,
            channel_number=channel_number,
            value=value,
        )
        await call_async("device/stream_set_digital_output", request)

    def set_all_digital_outputs(
            self,
            values: List[DigitalOutputAction]
    ) -> None:
        """
        Sets values for all digital output channels.

        Args:
            values: The type of action to perform on the channel.
        """
        request = dto.StreamSetAllDigitalOutputsRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            stream_id=self._stream_id,
            values=values,
        )
        call("device/stream_set_all_digital_outputs", request)

    async def set_all_digital_outputs_async(
            self,
            values: List[DigitalOutputAction]
    ) -> None:
        """
        Sets values for all digital output channels.

        Args:
            values: The type of action to perform on the channel.
        """
        request = dto.StreamSetAllDigitalOutputsRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            stream_id=self._stream_id,
            values=values,
        )
        await call_async("device/stream_set_all_digital_outputs", request)

    def set_digital_output_schedule(
            self,
            channel_number: int,
            value: DigitalOutputAction,
            future_value: DigitalOutputAction,
            delay: float,
            unit: TimeUnits = Units.NATIVE
    ) -> None:
        """
        Sets current and future value for the specified digital output channel.
        Requires at least Firmware 7.37.

        Args:
            channel_number: Channel number starting at 1.
            value: The type of action to perform immediately on the channel.
            future_value: The type of action to perform in the future on the channel.
            delay: Delay between setting current value and setting future value.
            unit: Units of time.
        """
        if delay <= 0:
            raise ValueError('Delay must be a positive value.')

        request = dto.StreamSetDigitalOutputScheduleRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            stream_id=self._stream_id,
            channel_number=channel_number,
            value=value,
            future_value=future_value,
            delay=delay,
            unit=unit,
        )
        call("device/stream_set_digital_output_schedule", request)

    async def set_digital_output_schedule_async(
            self,
            channel_number: int,
            value: DigitalOutputAction,
            future_value: DigitalOutputAction,
            delay: float,
            unit: TimeUnits = Units.NATIVE
    ) -> None:
        """
        Sets current and future value for the specified digital output channel.
        Requires at least Firmware 7.37.

        Args:
            channel_number: Channel number starting at 1.
            value: The type of action to perform immediately on the channel.
            future_value: The type of action to perform in the future on the channel.
            delay: Delay between setting current value and setting future value.
            unit: Units of time.
        """
        if delay <= 0:
            raise ValueError('Delay must be a positive value.')

        request = dto.StreamSetDigitalOutputScheduleRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            stream_id=self._stream_id,
            channel_number=channel_number,
            value=value,
            future_value=future_value,
            delay=delay,
            unit=unit,
        )
        await call_async("device/stream_set_digital_output_schedule", request)

    def set_all_digital_outputs_schedule(
            self,
            values: List[DigitalOutputAction],
            future_values: List[DigitalOutputAction],
            delay: float,
            unit: TimeUnits = Units.NATIVE
    ) -> None:
        """
        Sets current and future values for all digital output channels.
        Requires at least Firmware 7.37.

        Args:
            values: The type of actions to perform immediately on output channels.
            future_values: The type of actions to perform in the future on output channels.
            delay: Delay between setting current values and setting future values.
            unit: Units of time.
        """
        if delay <= 0:
            raise ValueError('Delay must be a positive value.')

        request = dto.StreamSetAllDigitalOutputsScheduleRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            stream_id=self._stream_id,
            values=values,
            future_values=future_values,
            delay=delay,
            unit=unit,
        )
        call("device/stream_set_all_digital_outputs_schedule", request)

    async def set_all_digital_outputs_schedule_async(
            self,
            values: List[DigitalOutputAction],
            future_values: List[DigitalOutputAction],
            delay: float,
            unit: TimeUnits = Units.NATIVE
    ) -> None:
        """
        Sets current and future values for all digital output channels.
        Requires at least Firmware 7.37.

        Args:
            values: The type of actions to perform immediately on output channels.
            future_values: The type of actions to perform in the future on output channels.
            delay: Delay between setting current values and setting future values.
            unit: Units of time.
        """
        if delay <= 0:
            raise ValueError('Delay must be a positive value.')

        request = dto.StreamSetAllDigitalOutputsScheduleRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            stream_id=self._stream_id,
            values=values,
            future_values=future_values,
            delay=delay,
            unit=unit,
        )
        await call_async("device/stream_set_all_digital_outputs_schedule", request)

    def set_analog_output(
            self,
            channel_number: int,
            value: float
    ) -> None:
        """
        Sets value for the specified analog output channel.

        Args:
            channel_number: Channel number starting at 1.
            value: Value to set the output channel voltage to.
        """
        request = dto.StreamSetAnalogOutputRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            stream_id=self._stream_id,
            channel_number=channel_number,
            value=value,
        )
        call("device/stream_set_analog_output", request)

    async def set_analog_output_async(
            self,
            channel_number: int,
            value: float
    ) -> None:
        """
        Sets value for the specified analog output channel.

        Args:
            channel_number: Channel number starting at 1.
            value: Value to set the output channel voltage to.
        """
        request = dto.StreamSetAnalogOutputRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            stream_id=self._stream_id,
            channel_number=channel_number,
            value=value,
        )
        await call_async("device/stream_set_analog_output", request)

    def set_all_analog_outputs(
            self,
            values: List[float]
    ) -> None:
        """
        Sets values for all analog output channels.

        Args:
            values: Voltage values to set the output channels to.
        """
        request = dto.StreamSetAllAnalogOutputsRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            stream_id=self._stream_id,
            values=values,
        )
        call("device/stream_set_all_analog_outputs", request)

    async def set_all_analog_outputs_async(
            self,
            values: List[float]
    ) -> None:
        """
        Sets values for all analog output channels.

        Args:
            values: Voltage values to set the output channels to.
        """
        request = dto.StreamSetAllAnalogOutputsRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            stream_id=self._stream_id,
            values=values,
        )
        await call_async("device/stream_set_all_analog_outputs", request)

    def set_analog_output_schedule(
            self,
            channel_number: int,
            value: float,
            future_value: float,
            delay: float,
            unit: TimeUnits = Units.NATIVE
    ) -> None:
        """
        Sets current and future value for the specified analog output channel.
        Requires at least Firmware 7.38.

        Args:
            channel_number: Channel number starting at 1.
            value: Value to set the output channel voltage to immediately.
            future_value: Value to set the output channel voltage to in the future.
            delay: Delay between setting current value and setting future value.
            unit: Units of time.
        """
        if delay <= 0:
            raise ValueError('Delay must be a positive value.')

        request = dto.StreamSetAnalogOutputScheduleRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            stream_id=self._stream_id,
            channel_number=channel_number,
            value=value,
            future_value=future_value,
            delay=delay,
            unit=unit,
        )
        call("device/stream_set_analog_output_schedule", request)

    async def set_analog_output_schedule_async(
            self,
            channel_number: int,
            value: float,
            future_value: float,
            delay: float,
            unit: TimeUnits = Units.NATIVE
    ) -> None:
        """
        Sets current and future value for the specified analog output channel.
        Requires at least Firmware 7.38.

        Args:
            channel_number: Channel number starting at 1.
            value: Value to set the output channel voltage to immediately.
            future_value: Value to set the output channel voltage to in the future.
            delay: Delay between setting current value and setting future value.
            unit: Units of time.
        """
        if delay <= 0:
            raise ValueError('Delay must be a positive value.')

        request = dto.StreamSetAnalogOutputScheduleRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            stream_id=self._stream_id,
            channel_number=channel_number,
            value=value,
            future_value=future_value,
            delay=delay,
            unit=unit,
        )
        await call_async("device/stream_set_analog_output_schedule", request)

    def set_all_analog_outputs_schedule(
            self,
            values: List[float],
            future_values: List[float],
            delay: float,
            unit: TimeUnits = Units.NATIVE
    ) -> None:
        """
        Sets current and future values for all analog output channels.
        Requires at least Firmware 7.38.

        Args:
            values: Voltage values to set the output channels to immediately.
            future_values: Voltage values to set the output channels to in the future.
            delay: Delay between setting current values and setting future values.
            unit: Units of time.
        """
        if delay <= 0:
            raise ValueError('Delay must be a positive value.')

        request = dto.StreamSetAllAnalogOutputsScheduleRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            stream_id=self._stream_id,
            values=values,
            future_values=future_values,
            delay=delay,
            unit=unit,
        )
        call("device/stream_set_all_analog_outputs_schedule", request)

    async def set_all_analog_outputs_schedule_async(
            self,
            values: List[float],
            future_values: List[float],
            delay: float,
            unit: TimeUnits = Units.NATIVE
    ) -> None:
        """
        Sets current and future values for all analog output channels.
        Requires at least Firmware 7.38.

        Args:
            values: Voltage values to set the output channels to immediately.
            future_values: Voltage values to set the output channels to in the future.
            delay: Delay between setting current values and setting future values.
            unit: Units of time.
        """
        if delay <= 0:
            raise ValueError('Delay must be a positive value.')

        request = dto.StreamSetAllAnalogOutputsScheduleRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            stream_id=self._stream_id,
            values=values,
            future_values=future_values,
            delay=delay,
            unit=unit,
        )
        await call_async("device/stream_set_all_analog_outputs_schedule", request)

    def cancel_digital_output_schedule(
            self,
            channel_number: int
    ) -> None:
        """
        Cancels a scheduled digital output action.
        Requires at least Firmware 7.37.

        Args:
            channel_number: Channel number starting at 1.
        """
        request = dto.StreamCancelOutputScheduleRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            analog=False,
            stream_id=self._stream_id,
            channel_number=channel_number,
        )
        call("device/stream_cancel_output_schedule", request)

    async def cancel_digital_output_schedule_async(
            self,
            channel_number: int
    ) -> None:
        """
        Cancels a scheduled digital output action.
        Requires at least Firmware 7.37.

        Args:
            channel_number: Channel number starting at 1.
        """
        request = dto.StreamCancelOutputScheduleRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            analog=False,
            stream_id=self._stream_id,
            channel_number=channel_number,
        )
        await call_async("device/stream_cancel_output_schedule", request)

    def cancel_all_digital_outputs_schedule(
            self,
            channels: List[bool] = []
    ) -> None:
        """
        Cancel all scheduled digital output actions.
        Requires at least Firmware 7.37.

        Args:
            channels: Optionally specify which channels to cancel.
                Array length must be empty or equal to the number of channels on device.
                Specifying "True" for a channel will cancel the scheduled digital output action for that channel.
        """
        request = dto.StreamCancelAllOutputsScheduleRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            analog=False,
            stream_id=self._stream_id,
            channels=channels,
        )
        call("device/stream_cancel_all_outputs_schedule", request)

    async def cancel_all_digital_outputs_schedule_async(
            self,
            channels: List[bool] = []
    ) -> None:
        """
        Cancel all scheduled digital output actions.
        Requires at least Firmware 7.37.

        Args:
            channels: Optionally specify which channels to cancel.
                Array length must be empty or equal to the number of channels on device.
                Specifying "True" for a channel will cancel the scheduled digital output action for that channel.
        """
        request = dto.StreamCancelAllOutputsScheduleRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            analog=False,
            stream_id=self._stream_id,
            channels=channels,
        )
        await call_async("device/stream_cancel_all_outputs_schedule", request)

    def cancel_analog_output_schedule(
            self,
            channel_number: int
    ) -> None:
        """
        Cancels a scheduled analog output value.
        Requires at least Firmware 7.38.

        Args:
            channel_number: Channel number starting at 1.
        """
        request = dto.StreamCancelOutputScheduleRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            analog=True,
            stream_id=self._stream_id,
            channel_number=channel_number,
        )
        call("device/stream_cancel_output_schedule", request)

    async def cancel_analog_output_schedule_async(
            self,
            channel_number: int
    ) -> None:
        """
        Cancels a scheduled analog output value.
        Requires at least Firmware 7.38.

        Args:
            channel_number: Channel number starting at 1.
        """
        request = dto.StreamCancelOutputScheduleRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            analog=True,
            stream_id=self._stream_id,
            channel_number=channel_number,
        )
        await call_async("device/stream_cancel_output_schedule", request)

    def cancel_all_analog_outputs_schedule(
            self,
            channels: List[bool] = []
    ) -> None:
        """
        Cancel all scheduled analog output actions.
        Requires at least Firmware 7.38.

        Args:
            channels: Optionally specify which channels to cancel.
                Array length must be empty or equal to the number of channels on device.
                Specifying "True" for a channel will cancel the scheduled analog output value for that channel.
        """
        request = dto.StreamCancelAllOutputsScheduleRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            analog=True,
            stream_id=self._stream_id,
            channels=channels,
        )
        call("device/stream_cancel_all_outputs_schedule", request)

    async def cancel_all_analog_outputs_schedule_async(
            self,
            channels: List[bool] = []
    ) -> None:
        """
        Cancel all scheduled analog output actions.
        Requires at least Firmware 7.38.

        Args:
            channels: Optionally specify which channels to cancel.
                Array length must be empty or equal to the number of channels on device.
                Specifying "True" for a channel will cancel the scheduled analog output value for that channel.
        """
        request = dto.StreamCancelAllOutputsScheduleRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            analog=True,
            stream_id=self._stream_id,
            channels=channels,
        )
        await call_async("device/stream_cancel_all_outputs_schedule", request)

    def wait_digital_input(
            self,
            channel_number: int,
            value: bool
    ) -> None:
        """
        Wait for a digital input channel to reach a given value.

        Args:
            channel_number: The number of the digital input channel.
                Channel numbers are numbered from one.
            value: The value that the stream should wait for.
        """
        request = dto.StreamWaitDigitalInputRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            stream_id=self._stream_id,
            channel_number=channel_number,
            value=value,
        )
        call("device/stream_wait_digital_input", request)

    async def wait_digital_input_async(
            self,
            channel_number: int,
            value: bool
    ) -> None:
        """
        Wait for a digital input channel to reach a given value.

        Args:
            channel_number: The number of the digital input channel.
                Channel numbers are numbered from one.
            value: The value that the stream should wait for.
        """
        request = dto.StreamWaitDigitalInputRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            stream_id=self._stream_id,
            channel_number=channel_number,
            value=value,
        )
        await call_async("device/stream_wait_digital_input", request)

    def wait_analog_input(
            self,
            channel_number: int,
            condition: str,
            value: float
    ) -> None:
        """
        Wait for the value of a analog input channel to reach a condition concerning a given value.

        Args:
            channel_number: The number of the analog input channel.
                Channel numbers are numbered from one.
            condition: A condition (e.g. <, <=, ==, !=).
            value: The value that the condition concerns, in Volts.
        """
        request = dto.StreamWaitAnalogInputRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            stream_id=self._stream_id,
            channel_number=channel_number,
            condition=condition,
            value=value,
        )
        call("device/stream_wait_analog_input", request)

    async def wait_analog_input_async(
            self,
            channel_number: int,
            condition: str,
            value: float
    ) -> None:
        """
        Wait for the value of a analog input channel to reach a condition concerning a given value.

        Args:
            channel_number: The number of the analog input channel.
                Channel numbers are numbered from one.
            condition: A condition (e.g. <, <=, ==, !=).
            value: The value that the condition concerns, in Volts.
        """
        request = dto.StreamWaitAnalogInputRequest(
            interface_id=self._device.connection.interface_id,
            device=self._device.device_address,
            stream_id=self._stream_id,
            channel_number=channel_number,
            condition=condition,
            value=value,
        )
        await call_async("device/stream_wait_analog_input", request)



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict, List
import decimal
from collections.abc import Iterable
import zaber_bson
from .servo_tuning_param import ServoTuningParam


@dataclass
class ParamsetInfo:
    """
    The raw parameters currently saved to a given paramset.
    """

    type: str
    """
    The tuning algorithm used for this axis.
    """

    version: int
    """
    The version of the tuning algorithm used for this axis.
    """

    params: List[ServoTuningParam]
    """
    The raw tuning parameters of this device.
    """

    @staticmethod
    def zero_values() -> 'ParamsetInfo':
        return ParamsetInfo(
            type="",
            version=0,
            params=[],
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'ParamsetInfo':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return ParamsetInfo.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'type': str(self.type or ''),
            'version': int(self.version),
            'params': [item.to_dict() for item in self.params] if self.params is not None else [],
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'ParamsetInfo':
        return ParamsetInfo(
            type=data.get('type'),  # type: ignore
            version=data.get('version'),  # type: ignore
            params=[ServoTuningParam.from_dict(item) for item in data.get('params')],  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.type is not None:
            if not isinstance(self.type, str):
                raise ValueError(f'Property "Type" of "ParamsetInfo" is not a string.')

        if self.version is None:
            raise ValueError(f'Property "Version" of "ParamsetInfo" is None.')

        if not isinstance(self.version, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Version" of "ParamsetInfo" is not a number.')

        if int(self.version) != self.version:
            raise ValueError(f'Property "Version" of "ParamsetInfo" is not integer value.')

        if self.params is not None:
            if not isinstance(self.params, Iterable):
                raise ValueError('Property "Params" of "ParamsetInfo" is not iterable.')

            for i, params_item in enumerate(self.params):
                if params_item is None:
                    raise ValueError(f'Item {i} in property "Params" of "ParamsetInfo" is None.')

                if not isinstance(params_item, ServoTuningParam):
                    raise ValueError(f'Item {i} in property "Params" of "ParamsetInfo" is not an instance of "ServoTuningParam".')

                params_item.validate()



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict, List
from collections.abc import Iterable
import zaber_bson
from .invalid_pvt_point import InvalidPvtPoint


@dataclass
class PvtExecutionExceptionData:
    """
    Contains additional data for PvtExecutionException.
    """

    error_flag: str
    """
    The error flag that caused the exception.
    """

    reason: str
    """
    The reason for the exception.
    """

    invalid_points: List[InvalidPvtPoint]
    """
    A list of points that cause the error (if applicable).
    """

    @staticmethod
    def zero_values() -> 'PvtExecutionExceptionData':
        return PvtExecutionExceptionData(
            error_flag="",
            reason="",
            invalid_points=[],
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'PvtExecutionExceptionData':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return PvtExecutionExceptionData.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'errorFlag': str(self.error_flag or ''),
            'reason': str(self.reason or ''),
            'invalidPoints': [item.to_dict() for item in self.invalid_points] if self.invalid_points is not None else [],
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'PvtExecutionExceptionData':
        return PvtExecutionExceptionData(
            error_flag=data.get('errorFlag'),  # type: ignore
            reason=data.get('reason'),  # type: ignore
            invalid_points=[InvalidPvtPoint.from_dict(item) for item in data.get('invalidPoints')],  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.error_flag is not None:
            if not isinstance(self.error_flag, str):
                raise ValueError(f'Property "ErrorFlag" of "PvtExecutionExceptionData" is not a string.')

        if self.reason is not None:
            if not isinstance(self.reason, str):
                raise ValueError(f'Property "Reason" of "PvtExecutionExceptionData" is not a string.')

        if self.invalid_points is not None:
            if not isinstance(self.invalid_points, Iterable):
                raise ValueError('Property "InvalidPoints" of "PvtExecutionExceptionData" is not iterable.')

            for i, invalid_points_item in enumerate(self.invalid_points):
                if invalid_points_item is None:
                    raise ValueError(f'Item {i} in property "InvalidPoints" of "PvtExecutionExceptionData" is None.')

                if not isinstance(invalid_points_item, InvalidPvtPoint):
                    raise ValueError(f'Item {i} in property "InvalidPoints" of "PvtExecutionExceptionData" is not an instance of "InvalidPvtPoint".')

                invalid_points_item.validate()



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from typing import Optional, List

from ..call import call, call_async, call_sync
from ..ascii import Device, Axis
from ..units import LengthUnits, Units

from ..dto import requests as dto
from ..dto.named_parameter import NamedParameter
from ..dto.microscopy.autofocus_status import AutofocusStatus


class Autofocus:
    """
    A generic autofocus device.
    """

    @property
    def provider_id(self) -> int:
        """
        The identification of external device providing the capability.
        """
        return self._provider_id

    @property
    def focus_axis(self) -> Axis:
        """
        The focus axis.
        """
        return self._focus_axis

    @property
    def objective_turret(self) -> Optional[Device]:
        """
        The objective turret device if the microscope has one.
        """
        return self._objective_turret

    def __init__(self, provider_id: int, focus_axis: Axis, objective_turret: Optional[Device]):
        """
        Creates instance of `Autofocus` based on the given provider id.
        """
        self._provider_id = provider_id
        self._focus_axis = focus_axis
        self._objective_turret = objective_turret

    def set_focus_zero(
            self
    ) -> None:
        """
        Sets the current focus to be target for the autofocus control loop.
        """
        request = dto.EmptyAutofocusRequest(
            provider_id=self.provider_id,
            interface_id=self.focus_axis.device.connection.interface_id,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
            turret_address=self.objective_turret.device_address if self.objective_turret else 0,
        )
        call("autofocus/set_zero", request)

    async def set_focus_zero_async(
            self
    ) -> None:
        """
        Sets the current focus to be target for the autofocus control loop.
        """
        request = dto.EmptyAutofocusRequest(
            provider_id=self.provider_id,
            interface_id=self.focus_axis.device.connection.interface_id,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
            turret_address=self.objective_turret.device_address if self.objective_turret else 0,
        )
        await call_async("autofocus/set_zero", request)

    def get_status(
            self
    ) -> AutofocusStatus:
        """
        Returns the status of the autofocus.

        Returns:
            The status of the autofocus.
        """
        request = dto.EmptyAutofocusRequest(
            provider_id=self.provider_id,
            interface_id=self.focus_axis.device.connection.interface_id,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
            turret_address=self.objective_turret.device_address if self.objective_turret else 0,
        )
        response = call(
            "autofocus/get_status",
            request,
            dto.AutofocusGetStatusResponse.from_binary)
        return response.status

    async def get_status_async(
            self
    ) -> AutofocusStatus:
        """
        Returns the status of the autofocus.

        Returns:
            The status of the autofocus.
        """
        request = dto.EmptyAutofocusRequest(
            provider_id=self.provider_id,
            interface_id=self.focus_axis.device.connection.interface_id,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
            turret_address=self.objective_turret.device_address if self.objective_turret else 0,
        )
        response = await call_async(
            "autofocus/get_status",
            request,
            dto.AutofocusGetStatusResponse.from_binary)
        return response.status

    def focus_once(
            self,
            scan: bool = False,
            timeout: int = -1
    ) -> None:
        """
        Moves the device until it's in focus.

        Args:
            scan: If true, the autofocus will approach from the limit moving until it's in range.
            timeout: Sets autofocus timeout duration in milliseconds.
        """
        request = dto.AutofocusFocusRequest(
            provider_id=self.provider_id,
            interface_id=self.focus_axis.device.connection.interface_id,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
            turret_address=self.objective_turret.device_address if self.objective_turret else 0,
            once=True,
            scan=scan,
            timeout=timeout,
        )
        call("autofocus/focus_once", request)

    async def focus_once_async(
            self,
            scan: bool = False,
            timeout: int = -1
    ) -> None:
        """
        Moves the device until it's in focus.

        Args:
            scan: If true, the autofocus will approach from the limit moving until it's in range.
            timeout: Sets autofocus timeout duration in milliseconds.
        """
        request = dto.AutofocusFocusRequest(
            provider_id=self.provider_id,
            interface_id=self.focus_axis.device.connection.interface_id,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
            turret_address=self.objective_turret.device_address if self.objective_turret else 0,
            once=True,
            scan=scan,
            timeout=timeout,
        )
        await call_async("autofocus/focus_once", request)

    def start_focus_loop(
            self
    ) -> None:
        """
        Moves the focus axis continuously maintaining focus.
        Starts the autofocus control loop.
        Note that the control loop may stop if the autofocus comes out of range or a movement error occurs.
        """
        request = dto.AutofocusFocusRequest(
            provider_id=self.provider_id,
            interface_id=self.focus_axis.device.connection.interface_id,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
            turret_address=self.objective_turret.device_address if self.objective_turret else 0,
        )
        call("autofocus/start_focus_loop", request)

    async def start_focus_loop_async(
            self
    ) -> None:
        """
        Moves the focus axis continuously maintaining focus.
        Starts the autofocus control loop.
        Note that the control loop may stop if the autofocus comes out of range or a movement error occurs.
        """
        request = dto.AutofocusFocusRequest(
            provider_id=self.provider_id,
            interface_id=self.focus_axis.device.connection.interface_id,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
            turret_address=self.objective_turret.device_address if self.objective_turret else 0,
        )
        await call_async("autofocus/start_focus_loop", request)

    def stop_focus_loop(
            self
    ) -> None:
        """
        Stops autofocus control loop.
        If the focus axis already stopped moving because of an error, an exception will be thrown.
        """
        request = dto.EmptyAutofocusRequest(
            provider_id=self.provider_id,
            interface_id=self.focus_axis.device.connection.interface_id,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
            turret_address=self.objective_turret.device_address if self.objective_turret else 0,
        )
        call("autofocus/stop_focus_loop", request)

    async def stop_focus_loop_async(
            self
    ) -> None:
        """
        Stops autofocus control loop.
        If the focus axis already stopped moving because of an error, an exception will be thrown.
        """
        request = dto.EmptyAutofocusRequest(
            provider_id=self.provider_id,
            interface_id=self.focus_axis.device.connection.interface_id,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
            turret_address=self.objective_turret.device_address if self.objective_turret else 0,
        )
        await call_async("autofocus/stop_focus_loop", request)

    def get_limit_min(
            self,
            unit: LengthUnits = Units.NATIVE
    ) -> float:
        """
        Gets the lower motion limit for the autofocus control loop.
        Gets motion.tracking.limit.min setting of the focus axis.

        Args:
            unit: The units of the limit.

        Returns:
            Limit value.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self.focus_axis.device.connection.interface_id,
            device=self.focus_axis.device.device_address,
            axis=self.focus_axis.axis_number,
            setting="motion.tracking.limit.min",
            unit=unit,
        )
        response = call(
            "device/get_setting",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    async def get_limit_min_async(
            self,
            unit: LengthUnits = Units.NATIVE
    ) -> float:
        """
        Gets the lower motion limit for the autofocus control loop.
        Gets motion.tracking.limit.min setting of the focus axis.

        Args:
            unit: The units of the limit.

        Returns:
            Limit value.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self.focus_axis.device.connection.interface_id,
            device=self.focus_axis.device.device_address,
            axis=self.focus_axis.axis_number,
            setting="motion.tracking.limit.min",
            unit=unit,
        )
        response = await call_async(
            "device/get_setting",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def get_limit_max(
            self,
            unit: LengthUnits = Units.NATIVE
    ) -> float:
        """
        Gets the upper motion limit for the autofocus control loop.
        Gets motion.tracking.limit.max setting of the focus axis.

        Args:
            unit: The units of the limit.

        Returns:
            Limit value.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self.focus_axis.device.connection.interface_id,
            device=self.focus_axis.device.device_address,
            axis=self.focus_axis.axis_number,
            setting="motion.tracking.limit.max",
            unit=unit,
        )
        response = call(
            "device/get_setting",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    async def get_limit_max_async(
            self,
            unit: LengthUnits = Units.NATIVE
    ) -> float:
        """
        Gets the upper motion limit for the autofocus control loop.
        Gets motion.tracking.limit.max setting of the focus axis.

        Args:
            unit: The units of the limit.

        Returns:
            Limit value.
        """
        request = dto.DeviceGetSettingRequest(
            interface_id=self.focus_axis.device.connection.interface_id,
            device=self.focus_axis.device.device_address,
            axis=self.focus_axis.axis_number,
            setting="motion.tracking.limit.max",
            unit=unit,
        )
        response = await call_async(
            "device/get_setting",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def set_limit_min(
            self,
            limit: float,
            unit: LengthUnits = Units.NATIVE
    ) -> None:
        """
        Sets the lower motion limit for the autofocus control loop.
        Use the limits to prevent the focus axis from crashing into the sample.
        Changes motion.tracking.limit.min setting of the focus axis.

        Args:
            limit: The lower limit of the focus axis.
            unit: The units of the limit.
        """
        request = dto.DeviceSetSettingRequest(
            interface_id=self.focus_axis.device.connection.interface_id,
            device=self.focus_axis.device.device_address,
            axis=self.focus_axis.axis_number,
            setting="motion.tracking.limit.min",
            value=limit,
            unit=unit,
        )
        call("device/set_setting", request)

    async def set_limit_min_async(
            self,
            limit: float,
            unit: LengthUnits = Units.NATIVE
    ) -> None:
        """
        Sets the lower motion limit for the autofocus control loop.
        Use the limits to prevent the focus axis from crashing into the sample.
        Changes motion.tracking.limit.min setting of the focus axis.

        Args:
            limit: The lower limit of the focus axis.
            unit: The units of the limit.
        """
        request = dto.DeviceSetSettingRequest(
            interface_id=self.focus_axis.device.connection.interface_id,
            device=self.focus_axis.device.device_address,
            axis=self.focus_axis.axis_number,
            setting="motion.tracking.limit.min",
            value=limit,
            unit=unit,
        )
        await call_async("device/set_setting", request)

    def set_limit_max(
            self,
            limit: float,
            unit: LengthUnits = Units.NATIVE
    ) -> None:
        """
        Sets the upper motion limit for the autofocus control loop.
        Use the limits to prevent the focus axis from crashing into the sample.
        Changes motion.tracking.limit.max setting of the focus axis.

        Args:
            limit: The upper limit of the focus axis.
            unit: The units of the limit.
        """
        request = dto.DeviceSetSettingRequest(
            interface_id=self.focus_axis.device.connection.interface_id,
            device=self.focus_axis.device.device_address,
            axis=self.focus_axis.axis_number,
            setting="motion.tracking.limit.max",
            value=limit,
            unit=unit,
        )
        call("device/set_setting", request)

    async def set_limit_max_async(
            self,
            limit: float,
            unit: LengthUnits = Units.NATIVE
    ) -> None:
        """
        Sets the upper motion limit for the autofocus control loop.
        Use the limits to prevent the focus axis from crashing into the sample.
        Changes motion.tracking.limit.max setting of the focus axis.

        Args:
            limit: The upper limit of the focus axis.
            unit: The units of the limit.
        """
        request = dto.DeviceSetSettingRequest(
            interface_id=self.focus_axis.device.connection.interface_id,
            device=self.focus_axis.device.device_address,
            axis=self.focus_axis.axis_number,
            setting="motion.tracking.limit.max",
            value=limit,
            unit=unit,
        )
        await call_async("device/set_setting", request)

    def synchronize_parameters(
            self
    ) -> None:
        """
        Typically, the control loop parameters and objective are kept synchronized by the library.
        If the parameters or current objective changes outside of the library, call this method to synchronize them.
        """
        request = dto.EmptyAutofocusRequest(
            provider_id=self.provider_id,
            interface_id=self.focus_axis.device.connection.interface_id,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
            turret_address=self.objective_turret.device_address if self.objective_turret else 0,
        )
        call("autofocus/sync_params", request)

    async def synchronize_parameters_async(
            self
    ) -> None:
        """
        Typically, the control loop parameters and objective are kept synchronized by the library.
        If the parameters or current objective changes outside of the library, call this method to synchronize them.
        """
        request = dto.EmptyAutofocusRequest(
            provider_id=self.provider_id,
            interface_id=self.focus_axis.device.connection.interface_id,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
            turret_address=self.objective_turret.device_address if self.objective_turret else 0,
        )
        await call_async("autofocus/sync_params", request)

    def set_objective_parameters(
            self,
            objective: int,
            parameters: List[NamedParameter]
    ) -> None:
        """
        Sets the parameters for the autofocus objective.
        Note that the method temporarily switches current objective to set the parameters.

        Args:
            objective: The objective (numbered from 1) to set the parameters for.
                If your microscope has only one objective, use value of 1.
            parameters: The parameters for the autofocus objective.
        """
        request = dto.AutofocusSetObjectiveParamsRequest(
            provider_id=self.provider_id,
            interface_id=self.focus_axis.device.connection.interface_id,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
            turret_address=self.objective_turret.device_address if self.objective_turret else 0,
            objective=objective,
            parameters=parameters,
        )
        call("autofocus/set_objective_params", request)

    async def set_objective_parameters_async(
            self,
            objective: int,
            parameters: List[NamedParameter]
    ) -> None:
        """
        Sets the parameters for the autofocus objective.
        Note that the method temporarily switches current objective to set the parameters.

        Args:
            objective: The objective (numbered from 1) to set the parameters for.
                If your microscope has only one objective, use value of 1.
            parameters: The parameters for the autofocus objective.
        """
        request = dto.AutofocusSetObjectiveParamsRequest(
            provider_id=self.provider_id,
            interface_id=self.focus_axis.device.connection.interface_id,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
            turret_address=self.objective_turret.device_address if self.objective_turret else 0,
            objective=objective,
            parameters=parameters,
        )
        await call_async("autofocus/set_objective_params", request)

    def get_objective_parameters(
            self,
            objective: int
    ) -> List[NamedParameter]:
        """
        Returns the parameters for the autofocus objective.

        Args:
            objective: The objective (numbered from 1) to get the parameters for.
                If your microscope has only one objective, use value of 1.
                Note that the method temporarily switches current objective to get the parameters.

        Returns:
            The parameters for the autofocus objective.
        """
        request = dto.AutofocusGetObjectiveParamsRequest(
            provider_id=self.provider_id,
            interface_id=self.focus_axis.device.connection.interface_id,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
            turret_address=self.objective_turret.device_address if self.objective_turret else 0,
            objective=objective,
        )
        response = call(
            "autofocus/get_objective_params",
            request,
            dto.AutofocusGetObjectiveParamsResponse.from_binary)
        return response.parameters

    async def get_objective_parameters_async(
            self,
            objective: int
    ) -> List[NamedParameter]:
        """
        Returns the parameters for the autofocus objective.

        Args:
            objective: The objective (numbered from 1) to get the parameters for.
                If your microscope has only one objective, use value of 1.
                Note that the method temporarily switches current objective to get the parameters.

        Returns:
            The parameters for the autofocus objective.
        """
        request = dto.AutofocusGetObjectiveParamsRequest(
            provider_id=self.provider_id,
            interface_id=self.focus_axis.device.connection.interface_id,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
            turret_address=self.objective_turret.device_address if self.objective_turret else 0,
            objective=objective,
        )
        response = await call_async(
            "autofocus/get_objective_params",
            request,
            dto.AutofocusGetObjectiveParamsResponse.from_binary)
        return response.parameters

    def __repr__(
            self
    ) -> str:
        """
        Returns a string that represents the autofocus.

        Returns:
            A string that represents the autofocus.
        """
        request = dto.EmptyAutofocusRequest(
            provider_id=self.provider_id,
            interface_id=self.focus_axis.device.connection.interface_id,
            focus_address=self.focus_axis.device.device_address,
            focus_axis=self.focus_axis.axis_number,
            turret_address=self.objective_turret.device_address if self.objective_turret else 0,
        )
        response = call_sync(
            "autofocus/to_string",
            request,
            dto.StringResponse.from_binary)
        return response.value



import typing

if typing.TYPE_CHECKING:
    from typing_extensions import Protocol
else:
    Protocol = object


class DtoObject(Protocol):
    """ Protocol for DTO objects. """

    def to_binary(self) -> bytes:
        """" Serializes to a binary. """

    def validate(self) -> None:
        """" Checks if the object is valid. """



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict, List
import decimal
from collections.abc import Iterable
import zaber_bson


@dataclass
class TriggerState:
    """
    The complete state of a trigger.
    """

    condition: str
    """
    The firing condition for a trigger.
    """

    actions: List[str]
    """
    The actions for a trigger.
    """

    enabled: bool
    """
    The enabled state for a trigger.
    """

    fires_total: int
    """
    The number of total fires for this trigger.
    A value of -1 indicates unlimited fires.
    """

    fires_remaining: int
    """
    The number of remaining fires for this trigger.
    A value of -1 indicates unlimited fires remaining.
    """

    @staticmethod
    def zero_values() -> 'TriggerState':
        return TriggerState(
            condition="",
            actions=[],
            enabled=False,
            fires_total=0,
            fires_remaining=0,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'TriggerState':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return TriggerState.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'condition': str(self.condition or ''),
            'actions': [str(item or '') for item in self.actions] if self.actions is not None else [],
            'enabled': bool(self.enabled),
            'firesTotal': int(self.fires_total),
            'firesRemaining': int(self.fires_remaining),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'TriggerState':
        return TriggerState(
            condition=data.get('condition'),  # type: ignore
            actions=data.get('actions'),  # type: ignore
            enabled=data.get('enabled'),  # type: ignore
            fires_total=data.get('firesTotal'),  # type: ignore
            fires_remaining=data.get('firesRemaining'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.condition is not None:
            if not isinstance(self.condition, str):
                raise ValueError(f'Property "Condition" of "TriggerState" is not a string.')

        if self.actions is not None:
            if not isinstance(self.actions, Iterable):
                raise ValueError('Property "Actions" of "TriggerState" is not iterable.')

            for i, actions_item in enumerate(self.actions):
                if actions_item is not None:
                    if not isinstance(actions_item, str):
                        raise ValueError(f'Item {i} in property "Actions" of "TriggerState" is not a string.')

        if self.fires_total is None:
            raise ValueError(f'Property "FiresTotal" of "TriggerState" is None.')

        if not isinstance(self.fires_total, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "FiresTotal" of "TriggerState" is not a number.')

        if int(self.fires_total) != self.fires_total:
            raise ValueError(f'Property "FiresTotal" of "TriggerState" is not integer value.')

        if self.fires_remaining is None:
            raise ValueError(f'Property "FiresRemaining" of "TriggerState" is None.')

        if not isinstance(self.fires_remaining, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "FiresRemaining" of "TriggerState" is not a number.')

        if int(self.fires_remaining) != self.fires_remaining:
            raise ValueError(f'Property "FiresRemaining" of "TriggerState" is not integer value.')



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from typing import Generator, List, Any, Optional, TypeVar
import asyncio
from reactivex.subject import ReplaySubject
from reactivex import operators as rxop, Observable

from ..call import call, call_async, call_sync
from ..convert_exception import convert_exception
from ..events import filter_events
from ..exceptions.motion_lib_exception import MotionLibException

from .device import Device
from .transport import Transport
from ..dto.ascii.response import Response
from ..dto.ascii.unknown_response_event import UnknownResponseEvent
from ..dto.ascii.alert_event import AlertEvent
from ..dto import requests as dto


TConnectionEvents = TypeVar(
    "TConnectionEvents",
    dto.UnknownResponseEventWrapper,
    dto.AlertEventWrapper,
    dto.DisconnectedEvent)


class Connection:
    """
    Class representing access to particular connection (serial port, TCP connection).
    """

    @property
    def unknown_response(self) -> Observable[UnknownResponseEvent]:
        """
        Event invoked when a response from a device cannot be matched to any known request.
        """
        return self._unknown_response

    @property
    def alert(self) -> Observable[AlertEvent]:
        """
        Event invoked when an alert is received from a device.
        """
        return self._alert

    @property
    def disconnected(self) -> Observable[MotionLibException]:
        """
        Event invoked when connection is interrupted or closed.
        """
        return self._disconnected

    DEFAULT_BAUD_RATE = 115200
    """
    Default baud rate for serial connections.
    """

    TCP_PORT_CHAIN = 55550
    """
    Commands sent over this port are forwarded to the device chain.
    The bandwidth may be limited as the commands are forwarded over a serial connection.
    """

    NETWORK_SHARE_PORT = 11421
    """
    Local area network share port.
    """

    TCP_PORT_DEVICE_ONLY = 55551
    """
    Commands send over this port are processed only by the device
    and not forwarded to the rest of the chain.
    Using this port typically makes the communication faster.
    """

    @property
    def interface_id(self) -> int:
        """
        The interface ID identifies this Connection instance with the underlying library.
        """
        return self._interface_id

    @property
    def default_request_timeout(self) -> int:
        """
        The default timeout, in milliseconds, for a device to respond to a request.
        Setting the timeout to a too low value may cause request timeout exceptions.
        The initial value is 1000 (one second).
        """
        return self.__retrieve_timeout()

    @default_request_timeout.setter
    def default_request_timeout(self, value: int) -> None:
        """
        The default timeout, in milliseconds, for a device to respond to a request.
        Setting the timeout to a too low value may cause request timeout exceptions.
        The initial value is 1000 (one second).
        """
        self.__change_timeout(value)

    @property
    def checksum_enabled(self) -> bool:
        """
        Controls whether outgoing messages contain checksum.
        """
        return self.__retrieve_checksum_enabled()

    @checksum_enabled.setter
    def checksum_enabled(self, value: bool) -> None:
        """
        Controls whether outgoing messages contain checksum.
        """
        self.__change_checksum_enabled(value)

    def __init__(self, interface_id: int):
        self._interface_id = interface_id
        self.__setup_events()

    @staticmethod
    def open_serial_port(
            port_name: str,
            baud_rate: int = DEFAULT_BAUD_RATE,
            direct: bool = False
    ) -> 'Connection':
        """
        Opens a serial port, if Zaber Launcher controls the port, the port will be opened through Zaber Launcher.
        Zaber Launcher allows sharing of the port between multiple applications,
        If port sharing is not desirable, use the `direct` parameter.

        Args:
            port_name: Name of the port to open.
            baud_rate: Optional baud rate (defaults to 115200).
            direct: If true will connect to the serial port directly,
                failing if the connection is already opened by a message router instance.

        Returns:
            An object representing the port.
        """
        request = dto.OpenInterfaceRequest(
            interface_type=dto.InterfaceType.SERIAL_PORT,
            port_name=port_name,
            baud_rate=baud_rate,
            reject_routed_connection=direct,
        )
        response = call(
            "interface/open",
            request,
            dto.OpenInterfaceResponse.from_binary)
        return Connection(response.interface_id)

    @staticmethod
    def open_serial_port_async(
            port_name: str,
            baud_rate: int = DEFAULT_BAUD_RATE,
            direct: bool = False
    ) -> 'AsyncConnectionOpener':
        """
        Opens a serial port, if Zaber Launcher controls the port, the port will be opened through Zaber Launcher.
        Zaber Launcher allows sharing of the port between multiple applications,
        If port sharing is not desirable, use the `direct` parameter.

        Args:
            port_name: Name of the port to open.
            baud_rate: Optional baud rate (defaults to 115200).
            direct: If true will connect to the serial port directly,
                failing if the connection is already opened by a message router instance.

        Returns:
            An object representing the port.
        """
        request = dto.OpenInterfaceRequest(
            interface_type=dto.InterfaceType.SERIAL_PORT,
            port_name=port_name,
            baud_rate=baud_rate,
            reject_routed_connection=direct,
        )
        return AsyncConnectionOpener(request)

    @staticmethod
    def open_tcp(
            host_name: str,
            port: int = TCP_PORT_CHAIN
    ) -> 'Connection':
        """
        Opens a TCP connection.

        Args:
            host_name: Hostname or IP address.
            port: Optional port number (defaults to 55550).

        Returns:
            An object representing the connection.
        """
        request = dto.OpenInterfaceRequest(
            interface_type=dto.InterfaceType.TCP,
            host_name=host_name,
            port=port,
        )
        response = call(
            "interface/open",
            request,
            dto.OpenInterfaceResponse.from_binary)
        return Connection(response.interface_id)

    @staticmethod
    def open_tcp_async(
            host_name: str,
            port: int = TCP_PORT_CHAIN
    ) -> 'AsyncConnectionOpener':
        """
        Opens a TCP connection.

        Args:
            host_name: Hostname or IP address.
            port: Optional port number (defaults to 55550).

        Returns:
            An object representing the connection.
        """
        request = dto.OpenInterfaceRequest(
            interface_type=dto.InterfaceType.TCP,
            host_name=host_name,
            port=port,
        )
        return AsyncConnectionOpener(request)

    @staticmethod
    def open_custom(
            transport: Transport
    ) -> 'Connection':
        """
        Opens a connection using a custom transport.

        Args:
            transport: The custom connection transport.

        Returns:
            An object representing the connection.
        """
        request = dto.OpenInterfaceRequest(
            interface_type=dto.InterfaceType.CUSTOM,
            transport=transport.transport_id,
        )
        response = call(
            "interface/open",
            request,
            dto.OpenInterfaceResponse.from_binary)
        return Connection(response.interface_id)

    @staticmethod
    def open_custom_async(
            transport: Transport
    ) -> 'AsyncConnectionOpener':
        """
        Opens a connection using a custom transport.

        Args:
            transport: The custom connection transport.

        Returns:
            An object representing the connection.
        """
        request = dto.OpenInterfaceRequest(
            interface_type=dto.InterfaceType.CUSTOM,
            transport=transport.transport_id,
        )
        return AsyncConnectionOpener(request)

    @staticmethod
    def open_iot(
            cloud_id: str,
            token: str = "unauthenticated",
            connection_name: Optional[str] = None,
            realm: Optional[str] = None,
            api: str = "https://api.zaber.io"
    ) -> 'Connection':
        """
        Opens a secured connection to a cloud connected device chain.
        Use this method to connect to devices on your account.

        Args:
            cloud_id: The cloud ID to connect to.
            token: The token to authenticate with. By default the connection will be unauthenticated.
            connection_name: The name of the connection to open.
                Can be left empty to default to the only connection present.
                Otherwise, use serial port name for serial port connection or hostname:port for TCP connection.
            realm: The realm to connect to.
                Can be left empty for the default account realm.
            api: The URL of the API to receive connection info from.

        Returns:
            An object representing the connection.
        """
        request = dto.OpenInterfaceRequest(
            interface_type=dto.InterfaceType.IOT,
            cloud_id=cloud_id,
            token=token,
            connection_name=connection_name,
            realm=realm,
            api=api,
        )
        response = call(
            "interface/open",
            request,
            dto.OpenInterfaceResponse.from_binary)
        return Connection(response.interface_id)

    @staticmethod
    def open_iot_async(
            cloud_id: str,
            token: str = "unauthenticated",
            connection_name: Optional[str] = None,
            realm: Optional[str] = None,
            api: str = "https://api.zaber.io"
    ) -> 'AsyncConnectionOpener':
        """
        Opens a secured connection to a cloud connected device chain.
        Use this method to connect to devices on your account.

        Args:
            cloud_id: The cloud ID to connect to.
            token: The token to authenticate with. By default the connection will be unauthenticated.
            connection_name: The name of the connection to open.
                Can be left empty to default to the only connection present.
                Otherwise, use serial port name for serial port connection or hostname:port for TCP connection.
            realm: The realm to connect to.
                Can be left empty for the default account realm.
            api: The URL of the API to receive connection info from.

        Returns:
            An object representing the connection.
        """
        request = dto.OpenInterfaceRequest(
            interface_type=dto.InterfaceType.IOT,
            cloud_id=cloud_id,
            token=token,
            connection_name=connection_name,
            realm=realm,
            api=api,
        )
        return AsyncConnectionOpener(request)

    @staticmethod
    def open_network_share(
            host_name: str,
            port: int = NETWORK_SHARE_PORT,
            connection_name: Optional[str] = None
    ) -> 'Connection':
        """
        Opens a connection to Zaber Launcher in your Local Area Network.
        The connection is not secured.

        Args:
            host_name: Hostname or IP address.
            port: Port number.
            connection_name: The name of the connection to open.
                Can be left empty to default to the only connection present.
                Otherwise, use serial port name for serial port connection or hostname:port for TCP connection.

        Returns:
            An object representing the connection.
        """
        request = dto.OpenInterfaceRequest(
            interface_type=dto.InterfaceType.NETWORK_SHARE,
            host_name=host_name,
            port=port,
            connection_name=connection_name,
        )
        response = call(
            "interface/open",
            request,
            dto.OpenInterfaceResponse.from_binary)
        return Connection(response.interface_id)

    @staticmethod
    def open_network_share_async(
            host_name: str,
            port: int = NETWORK_SHARE_PORT,
            connection_name: Optional[str] = None
    ) -> 'AsyncConnectionOpener':
        """
        Opens a connection to Zaber Launcher in your Local Area Network.
        The connection is not secured.

        Args:
            host_name: Hostname or IP address.
            port: Port number.
            connection_name: The name of the connection to open.
                Can be left empty to default to the only connection present.
                Otherwise, use serial port name for serial port connection or hostname:port for TCP connection.

        Returns:
            An object representing the connection.
        """
        request = dto.OpenInterfaceRequest(
            interface_type=dto.InterfaceType.NETWORK_SHARE,
            host_name=host_name,
            port=port,
            connection_name=connection_name,
        )
        return AsyncConnectionOpener(request)

    def generic_command(
            self,
            command: str,
            device: int = 0,
            axis: int = 0,
            check_errors: bool = True,
            timeout: int = 0
    ) -> Response:
        """
        Sends a generic ASCII command to this connection.
        For more information refer to the [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command: Command and its parameters.
            device: Optional device address to send the command to.
            axis: Optional axis number to send the command to.
            check_errors: Controls whether to throw an exception when the device rejects the command.
            timeout: The timeout, in milliseconds, for a device to respond to the command.
                Overrides the connection default request timeout.

        Returns:
            A response to the command.
        """
        request = dto.GenericCommandRequest(
            interface_id=self.interface_id,
            command=command,
            device=device,
            axis=axis,
            check_errors=check_errors,
            timeout=timeout,
        )
        response = call(
            "interface/generic_command",
            request,
            Response.from_binary)
        return response

    async def generic_command_async(
            self,
            command: str,
            device: int = 0,
            axis: int = 0,
            check_errors: bool = True,
            timeout: int = 0
    ) -> Response:
        """
        Sends a generic ASCII command to this connection.
        For more information refer to the [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command: Command and its parameters.
            device: Optional device address to send the command to.
            axis: Optional axis number to send the command to.
            check_errors: Controls whether to throw an exception when the device rejects the command.
            timeout: The timeout, in milliseconds, for a device to respond to the command.
                Overrides the connection default request timeout.

        Returns:
            A response to the command.
        """
        request = dto.GenericCommandRequest(
            interface_id=self.interface_id,
            command=command,
            device=device,
            axis=axis,
            check_errors=check_errors,
            timeout=timeout,
        )
        response = await call_async(
            "interface/generic_command",
            request,
            Response.from_binary)
        return response

    def generic_command_no_response(
            self,
            command: str,
            device: int = 0,
            axis: int = 0
    ) -> None:
        """
        Sends a generic ASCII command to this connection without expecting a response and without adding a message ID.
        For more information refer to the [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command: Command and its parameters.
            device: Optional device address to send the command to.
                Specifying -1 omits the number completely.
            axis: Optional axis number to send the command to.
                Specifying -1 omits the number completely.
        """
        request = dto.GenericCommandRequest(
            interface_id=self.interface_id,
            command=command,
            device=device,
            axis=axis,
        )
        call("interface/generic_command_no_response", request)

    async def generic_command_no_response_async(
            self,
            command: str,
            device: int = 0,
            axis: int = 0
    ) -> None:
        """
        Sends a generic ASCII command to this connection without expecting a response and without adding a message ID.
        For more information refer to the [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command: Command and its parameters.
            device: Optional device address to send the command to.
                Specifying -1 omits the number completely.
            axis: Optional axis number to send the command to.
                Specifying -1 omits the number completely.
        """
        request = dto.GenericCommandRequest(
            interface_id=self.interface_id,
            command=command,
            device=device,
            axis=axis,
        )
        await call_async("interface/generic_command_no_response", request)

    def generic_command_multi_response(
            self,
            command: str,
            device: int = 0,
            axis: int = 0,
            check_errors: bool = True,
            timeout: int = 0
    ) -> List[Response]:
        """
        Sends a generic ASCII command to this connection and expect multiple responses,
        either from one device or from many devices.
        Responses are returned in order of arrival.
        For more information refer to the [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command: Command and its parameters.
            device: Optional device address to send the command to.
            axis: Optional axis number to send the command to.
            check_errors: Controls whether to throw an exception when a device rejects the command.
            timeout: The timeout, in milliseconds, for a device to respond to the command.
                Overrides the connection default request timeout.

        Returns:
            All responses to the command.
        """
        request = dto.GenericCommandRequest(
            interface_id=self.interface_id,
            command=command,
            device=device,
            axis=axis,
            check_errors=check_errors,
            timeout=timeout,
        )
        response = call(
            "interface/generic_command_multi_response",
            request,
            dto.GenericCommandResponseCollection.from_binary)
        return response.responses

    async def generic_command_multi_response_async(
            self,
            command: str,
            device: int = 0,
            axis: int = 0,
            check_errors: bool = True,
            timeout: int = 0
    ) -> List[Response]:
        """
        Sends a generic ASCII command to this connection and expect multiple responses,
        either from one device or from many devices.
        Responses are returned in order of arrival.
        For more information refer to the [ASCII Protocol Manual](https://www.zaber.com/protocol-manual#topic_commands).

        Args:
            command: Command and its parameters.
            device: Optional device address to send the command to.
            axis: Optional axis number to send the command to.
            check_errors: Controls whether to throw an exception when a device rejects the command.
            timeout: The timeout, in milliseconds, for a device to respond to the command.
                Overrides the connection default request timeout.

        Returns:
            All responses to the command.
        """
        request = dto.GenericCommandRequest(
            interface_id=self.interface_id,
            command=command,
            device=device,
            axis=axis,
            check_errors=check_errors,
            timeout=timeout,
        )
        response = await call_async(
            "interface/generic_command_multi_response",
            request,
            dto.GenericCommandResponseCollection.from_binary)
        return response.responses

    def enable_alerts(
            self
    ) -> None:
        """
        Enables alerts for all devices on the connection.
        This will change the "comm.alert" setting to 1 on all supported devices.
        """
        request = dto.GenericCommandRequest(
            interface_id=self.interface_id,
            command="set comm.alert 1",
        )
        call("interface/generic_command_no_response", request)

    async def enable_alerts_async(
            self
    ) -> None:
        """
        Enables alerts for all devices on the connection.
        This will change the "comm.alert" setting to 1 on all supported devices.
        """
        request = dto.GenericCommandRequest(
            interface_id=self.interface_id,
            command="set comm.alert 1",
        )
        await call_async("interface/generic_command_no_response", request)

    def disable_alerts(
            self
    ) -> None:
        """
        Disables alerts for all devices on the connection.
        This will change the "comm.alert" setting to 0 on all supported devices.
        """
        request = dto.GenericCommandRequest(
            interface_id=self.interface_id,
            command="set comm.alert 0",
        )
        call("interface/generic_command_no_response", request)

    async def disable_alerts_async(
            self
    ) -> None:
        """
        Disables alerts for all devices on the connection.
        This will change the "comm.alert" setting to 0 on all supported devices.
        """
        request = dto.GenericCommandRequest(
            interface_id=self.interface_id,
            command="set comm.alert 0",
        )
        await call_async("interface/generic_command_no_response", request)

    def reset_ids(
            self
    ) -> None:
        """
        Resets ASCII protocol message IDs. Only for testing purposes.
        """
        request = dto.InterfaceEmptyRequest(
            interface_id=self.interface_id,
        )
        call_sync("interface/reset_ids", request)

    def close(
            self
    ) -> None:
        """
        Close the connection.
        """
        request = dto.InterfaceEmptyRequest(
            interface_id=self.interface_id,
        )
        call("interface/close", request)

    async def close_async(
            self
    ) -> None:
        """
        Close the connection.
        """
        request = dto.InterfaceEmptyRequest(
            interface_id=self.interface_id,
        )
        await call_async("interface/close", request)

    def get_device(
            self,
            device_address: int
    ) -> Device:
        """
        Gets a Device class instance which allows you to control a particular device on this connection.
        Devices are numbered from 1.

        Args:
            device_address: Address of device intended to control. Address is configured for each device.

        Returns:
            Device instance.
        """
        if device_address <= 0:
            raise ValueError('Invalid value; physical devices are numbered from 1.')

        return Device(self, device_address)

    def renumber_devices(
            self,
            first_address: int = 1
    ) -> int:
        """
        Renumbers devices present on this connection. After renumbering, devices need to be identified again.

        Args:
            first_address: This is the address that the device closest to the computer is given.
                Remaining devices are numbered consecutively.

        Returns:
            Total number of devices that responded to the renumber.
        """
        if first_address <= 0:
            raise ValueError('Invalid value; device addresses are numbered from 1.')

        request = dto.RenumberRequest(
            interface_id=self.interface_id,
            address=first_address,
        )
        response = call(
            "device/renumber_all",
            request,
            dto.IntResponse.from_binary)
        return response.value

    async def renumber_devices_async(
            self,
            first_address: int = 1
    ) -> int:
        """
        Renumbers devices present on this connection. After renumbering, devices need to be identified again.

        Args:
            first_address: This is the address that the device closest to the computer is given.
                Remaining devices are numbered consecutively.

        Returns:
            Total number of devices that responded to the renumber.
        """
        if first_address <= 0:
            raise ValueError('Invalid value; device addresses are numbered from 1.')

        request = dto.RenumberRequest(
            interface_id=self.interface_id,
            address=first_address,
        )
        response = await call_async(
            "device/renumber_all",
            request,
            dto.IntResponse.from_binary)
        return response.value

    def detect_devices(
            self,
            identify_devices: bool = True
    ) -> List[Device]:
        """
        Attempts to detect any devices present on this connection.

        Args:
            identify_devices: Determines whether device identification should be performed as well.

        Returns:
            Array of detected devices.
        """
        request = dto.DeviceDetectRequest(
            interface_id=self.interface_id,
            identify_devices=identify_devices,
        )
        response = call(
            "device/detect",
            request,
            dto.DeviceDetectResponse.from_binary)
        return list(map(self.get_device, response.devices))

    async def detect_devices_async(
            self,
            identify_devices: bool = True
    ) -> List[Device]:
        """
        Attempts to detect any devices present on this connection.

        Args:
            identify_devices: Determines whether device identification should be performed as well.

        Returns:
            Array of detected devices.
        """
        request = dto.DeviceDetectRequest(
            interface_id=self.interface_id,
            identify_devices=identify_devices,
        )
        response = await call_async(
            "device/detect",
            request,
            dto.DeviceDetectResponse.from_binary)
        return list(map(self.get_device, response.devices))

    def forget_devices(
            self,
            except_devices: List[int] = []
    ) -> None:
        """
        Forgets all the information associated with devices on the connection.
        Useful when devices are removed from the chain indefinitely.

        Args:
            except_devices: Addresses of devices that should not be forgotten.
        """
        request = dto.ForgetDevicesRequest(
            interface_id=self.interface_id,
            except_devices=except_devices,
        )
        call_sync("device/forget", request)

    def stop_all(
            self,
            wait_until_idle: bool = True
    ) -> List[int]:
        """
        Stops all of the devices on this connection.

        Args:
            wait_until_idle: Determines whether the function should return immediately
                or wait until the devices are stopped.

        Returns:
            The addresses of the devices that were stopped by this command.
        """
        request = dto.DeviceOnAllRequest(
            interface_id=self.interface_id,
            wait_until_idle=wait_until_idle,
        )
        response = call(
            "device/stop_all",
            request,
            dto.DeviceOnAllResponse.from_binary)
        return response.device_addresses

    async def stop_all_async(
            self,
            wait_until_idle: bool = True
    ) -> List[int]:
        """
        Stops all of the devices on this connection.

        Args:
            wait_until_idle: Determines whether the function should return immediately
                or wait until the devices are stopped.

        Returns:
            The addresses of the devices that were stopped by this command.
        """
        request = dto.DeviceOnAllRequest(
            interface_id=self.interface_id,
            wait_until_idle=wait_until_idle,
        )
        response = await call_async(
            "device/stop_all",
            request,
            dto.DeviceOnAllResponse.from_binary)
        return response.device_addresses

    def home_all(
            self,
            wait_until_idle: bool = True
    ) -> List[int]:
        """
        Homes all of the devices on this connection.

        Args:
            wait_until_idle: Determines whether the function should return immediately
                or wait until the devices are homed.

        Returns:
            The addresses of the devices that were homed by this command.
        """
        request = dto.DeviceOnAllRequest(
            interface_id=self.interface_id,
            wait_until_idle=wait_until_idle,
        )
        response = call(
            "device/home_all",
            request,
            dto.DeviceOnAllResponse.from_binary)
        return response.device_addresses

    async def home_all_async(
            self,
            wait_until_idle: bool = True
    ) -> List[int]:
        """
        Homes all of the devices on this connection.

        Args:
            wait_until_idle: Determines whether the function should return immediately
                or wait until the devices are homed.

        Returns:
            The addresses of the devices that were homed by this command.
        """
        request = dto.DeviceOnAllRequest(
            interface_id=self.interface_id,
            wait_until_idle=wait_until_idle,
        )
        response = await call_async(
            "device/home_all",
            request,
            dto.DeviceOnAllResponse.from_binary)
        return response.device_addresses

    def __repr__(
            self
    ) -> str:
        """
        Returns a string that represents the connection.

        Returns:
            A string that represents the connection.
        """
        request = dto.InterfaceEmptyRequest(
            interface_id=self.interface_id,
        )
        response = call_sync(
            "interface/to_string",
            request,
            dto.StringResponse.from_binary)
        return response.value

    def __retrieve_timeout(
            self
    ) -> int:
        """
        Returns default request timeout.

        Returns:
            Default request timeout.
        """
        request = dto.InterfaceEmptyRequest(
            interface_id=self.interface_id,
        )
        response = call_sync(
            "interface/get_timeout",
            request,
            dto.IntResponse.from_binary)
        return response.value

    def __change_timeout(
            self,
            timeout: int
    ) -> None:
        """
        Sets default request timeout.

        Args:
            timeout: Default request timeout.
        """
        request = dto.SetInterfaceTimeoutRequest(
            interface_id=self.interface_id,
            timeout=timeout,
        )
        call_sync("interface/set_timeout", request)

    def __retrieve_checksum_enabled(
            self
    ) -> bool:
        """
        Returns checksum enabled.

        Returns:
            Checksum enabled.
        """
        request = dto.InterfaceEmptyRequest(
            interface_id=self.interface_id,
        )
        response = call_sync(
            "interface/get_checksum_enabled",
            request,
            dto.BoolResponse.from_binary)
        return response.value

    def __change_checksum_enabled(
            self,
            is_enabled: bool
    ) -> None:
        """
        Sets checksum enabled.

        Args:
            is_enabled: Checksum enabled.
        """
        request = dto.SetInterfaceChecksumEnabledRequest(
            interface_id=self.interface_id,
            is_enabled=is_enabled,
        )
        call_sync("interface/set_checksum_enabled", request)

    def __enter__(self) -> 'Connection':
        """ __enter__ """
        return self

    def __exit__(self, _type: Any, _value: Any, _traceback: Any) -> None:
        """ __exit__ """
        self.close()

    def __setup_events(self) -> None:
        def filter_connection_event(
            data: TConnectionEvents,
        ) -> bool:
            return data.interface_id == self._interface_id

        self._disconnected = ReplaySubject[MotionLibException]()  # terminates all the events

        unknown_response = filter_events('interface/unknown_response', dto.UnknownResponseEventWrapper)
        self._unknown_response = unknown_response.pipe(
            rxop.filter(filter_connection_event),
            rxop.take_until(self.disconnected),
            rxop.map(map_unknown_response)
        )

        alert = filter_events('interface/alert', dto.AlertEventWrapper)
        self._alert = alert.pipe(
            rxop.filter(filter_connection_event),
            rxop.take_until(self.disconnected),
            rxop.map(map_alert)
        )

        disconnected = filter_events('interface/disconnected', dto.DisconnectedEvent)
        disconnected.pipe(
            rxop.filter(filter_connection_event),
            rxop.take(1),
            rxop.map(map_disconnect)
        ).subscribe(self._disconnected)


class AsyncConnectionOpener:
    '''Provides a connection in an asynchronous context using `await` or `async with`'''
    def __init__(self, request: dto.OpenInterfaceRequest) -> None:
        self._request = request
        self._resource: Optional[Connection] = None

    async def _create_resource(self) -> Connection:
        task = asyncio.ensure_future(call_async(
            "interface/open",
            self._request,
            dto.OpenInterfaceResponse.from_binary))

        try:
            response = await asyncio.shield(task)
        except asyncio.CancelledError:
            async def cancel() -> None:
                try:
                    response = await task
                    await Connection(response.interface_id).close_async()
                except MotionLibException:
                    pass

            asyncio.ensure_future(cancel())
            raise

        return Connection(response.interface_id)

    def __await__(self) -> Generator[Any, None, 'Connection']:
        return self._create_resource().__await__()

    async def __aenter__(self) -> 'Connection':
        self._resource = await self._create_resource()
        return self._resource

    async def __aexit__(self, exc_type: Any, exc: Any, trace: Any) -> None:
        if self._resource is not None:
            await self._resource.close_async()


def map_unknown_response(event: dto.UnknownResponseEventWrapper) -> UnknownResponseEvent:
    return event.unknown_response


def map_alert(event: dto.AlertEventWrapper) -> AlertEvent:
    return event.alert


def map_disconnect(event: dto.DisconnectedEvent) -> MotionLibException:
    return convert_exception(event.error_type, event.error_message)



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from .command_failed_exception import CommandFailedException


class BadCommandException(CommandFailedException):
    """
    Thrown when a device receives an invalid command.
    """



# This file is generated. Do not modify by hand.
from enum import Enum


class ProcessControllerSourceSensor(Enum):
    """
    Servo Tuning Parameter Set to target.
    """

    THERMISTOR = 10
    ANALOG_INPUT = 20



# ===== THIS FILE IS GENERATED FROM A TEMPLATE ===== #
# ============== DO NOT EDIT DIRECTLY ============== #

from typing import List, Optional
from ..call import call, call_async, call_sync

from ..dto import requests as dto
from ..units import LengthUnits, VelocityUnits
from ..ascii import Device
from ..dto.gcode.device_definition import DeviceDefinition
from ..dto.gcode.translator_config import TranslatorConfig
from ..dto.gcode.translate_result import TranslateResult


class OfflineTranslator:
    """
    Represents an offline G-Code translator.
    It allows to translate G-Code blocks to Zaber ASCII protocol stream commands.
    This translator does not need a connected device to perform translation.
    Requires at least Firmware 7.11.
    """

    @property
    def translator_id(self) -> int:
        """
        The ID of the translator that serves to identify native resources.
        """
        return self._translator_id

    @property
    def coordinate_system(self) -> str:
        """
        Current coordinate system.
        """
        return self.__get_current_coordinate_system()

    def __init__(self, translator_id: int):
        self._translator_id = translator_id

    @staticmethod
    def setup(
            definition: DeviceDefinition,
            config: Optional[TranslatorConfig] = None
    ) -> 'OfflineTranslator':
        """
        Sets up translator from provided device definition and configuration.

        Args:
            definition: Definition of device and its peripherals.
                The definition must match a device that later performs the commands.
            config: Configuration of the translator.

        Returns:
            New instance of translator.
        """
        request = dto.TranslatorCreateRequest(
            definition=definition,
            config=config,
        )
        response = call(
            "gcode/create",
            request,
            dto.TranslatorCreateResponse.from_binary)
        return OfflineTranslator(response.translator_id)

    @staticmethod
    async def setup_async(
            definition: DeviceDefinition,
            config: Optional[TranslatorConfig] = None
    ) -> 'OfflineTranslator':
        """
        Sets up translator from provided device definition and configuration.

        Args:
            definition: Definition of device and its peripherals.
                The definition must match a device that later performs the commands.
            config: Configuration of the translator.

        Returns:
            New instance of translator.
        """
        request = dto.TranslatorCreateRequest(
            definition=definition,
            config=config,
        )
        response = await call_async(
            "gcode/create",
            request,
            dto.TranslatorCreateResponse.from_binary)
        return OfflineTranslator(response.translator_id)

    @staticmethod
    def setup_from_device(
            device: Device,
            axes: List[int],
            config: Optional[TranslatorConfig] = None
    ) -> 'OfflineTranslator':
        """
        Sets up an offline translator from provided device, axes, and configuration.

        Args:
            device: Device that later performs the command streaming.
            axes: Axis numbers that are later used to setup the stream.
                For a lockstep group specify only the first axis of the group.
            config: Configuration of the translator.

        Returns:
            New instance of translator.
        """
        request = dto.TranslatorCreateFromDeviceRequest(
            interface_id=device.connection.interface_id,
            device=device.device_address,
            axes=axes,
            config=config,
        )
        response = call(
            "gcode/create_from_device",
            request,
            dto.TranslatorCreateResponse.from_binary)
        return OfflineTranslator(response.translator_id)

    @staticmethod
    async def setup_from_device_async(
            device: Device,
            axes: List[int],
            config: Optional[TranslatorConfig] = None
    ) -> 'OfflineTranslator':
        """
        Sets up an offline translator from provided device, axes, and configuration.

        Args:
            device: Device that later performs the command streaming.
            axes: Axis numbers that are later used to setup the stream.
                For a lockstep group specify only the first axis of the group.
            config: Configuration of the translator.

        Returns:
            New instance of translator.
        """
        request = dto.TranslatorCreateFromDeviceRequest(
            interface_id=device.connection.interface_id,
            device=device.device_address,
            axes=axes,
            config=config,
        )
        response = await call_async(
            "gcode/create_from_device",
            request,
            dto.TranslatorCreateResponse.from_binary)
        return OfflineTranslator(response.translator_id)

    def translate(
            self,
            block: str
    ) -> TranslateResult:
        """
        Translates a single block (line) of G-code.

        Args:
            block: Block (line) of G-code.

        Returns:
            Result of translation containing the stream commands.
        """
        request = dto.TranslatorTranslateRequest(
            translator_id=self.translator_id,
            block=block,
        )
        response = call_sync(
            "gcode/translate",
            request,
            TranslateResult.from_binary)
        return response

    def flush(
            self
    ) -> List[str]:
        """
        Flushes the remaining stream commands waiting in optimization buffer.
        The flush is also performed by M2 and M30 codes.

        Returns:
            The remaining stream commands.
        """
        request = dto.TranslatorEmptyRequest(
            translator_id=self.translator_id,
        )
        response = call_sync(
            "gcode/flush",
            request,
            dto.TranslatorFlushResponse.from_binary)
        return response.commands

    def set_traverse_rate(
            self,
            traverse_rate: float,
            unit: VelocityUnits
    ) -> None:
        """
        Sets the speed at which the device moves when traversing (G0).

        Args:
            traverse_rate: The traverse rate.
            unit: Units of the traverse rate.
        """
        request = dto.TranslatorSetTraverseRateRequest(
            translator_id=self.translator_id,
            traverse_rate=traverse_rate,
            unit=unit,
        )
        call_sync("gcode/set_traverse_rate", request)

    def set_axis_position(
            self,
            axis: str,
            position: float,
            unit: LengthUnits
    ) -> None:
        """
        Sets position of translator's axis.
        Use this method to set position after performing movement outside of the translator.
        This method does not cause any movement.

        Args:
            axis: Letter of the axis.
            position: The position.
            unit: Units of position.
        """
        request = dto.TranslatorSetAxisPositionRequest(
            translator_id=self.translator_id,
            axis=axis,
            position=position,
            unit=unit,
        )
        call_sync("gcode/set_axis_position", request)

    def get_axis_position(
            self,
            axis: str,
            unit: LengthUnits
    ) -> float:
        """
        Gets position of translator's axis.
        This method does not query device but returns value from translator's state.

        Args:
            axis: Letter of the axis.
            unit: Units of position.

        Returns:
            Position of translator's axis.
        """
        request = dto.TranslatorGetAxisPositionRequest(
            translator_id=self.translator_id,
            axis=axis,
            unit=unit,
        )
        response = call_sync(
            "gcode/get_axis_position",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def set_axis_home_position(
            self,
            axis: str,
            position: float,
            unit: LengthUnits
    ) -> None:
        """
        Sets the home position of translator's axis.
        This position is used by G28.

        Args:
            axis: Letter of the axis.
            position: The home position.
            unit: Units of position.
        """
        request = dto.TranslatorSetAxisPositionRequest(
            translator_id=self.translator_id,
            axis=axis,
            position=position,
            unit=unit,
        )
        call_sync("gcode/set_axis_home", request)

    def set_axis_secondary_home_position(
            self,
            axis: str,
            position: float,
            unit: LengthUnits
    ) -> None:
        """
        Sets the secondary home position of translator's axis.
        This position is used by G30.

        Args:
            axis: Letter of the axis.
            position: The home position.
            unit: Units of position.
        """
        request = dto.TranslatorSetAxisPositionRequest(
            translator_id=self.translator_id,
            axis=axis,
            position=position,
            unit=unit,
        )
        call_sync("gcode/set_axis_secondary_home", request)

    def get_axis_coordinate_system_offset(
            self,
            coordinate_system: str,
            axis: str,
            unit: LengthUnits
    ) -> float:
        """
        Gets offset of an axis in a given coordinate system.

        Args:
            coordinate_system: Coordinate system (e.g. G54).
            axis: Letter of the axis.
            unit: Units of position.

        Returns:
            Offset in translator units of the axis.
        """
        request = dto.TranslatorGetAxisOffsetRequest(
            translator_id=self.translator_id,
            coordinate_system=coordinate_system,
            axis=axis,
            unit=unit,
        )
        response = call_sync(
            "gcode/get_axis_offset",
            request,
            dto.DoubleResponse.from_binary)
        return response.value

    def reset_after_stream_error(
            self
    ) -> None:
        """
        Resets internal state after device rejected generated command.
        Axis positions become uninitialized.
        """
        request = dto.TranslatorEmptyRequest(
            translator_id=self.translator_id,
        )
        call_sync("gcode/reset_after_stream_error", request)

    def set_feed_rate_override(
            self,
            coefficient: float
    ) -> None:
        """
        Allows to scale feed rate of the translated code by a coefficient.

        Args:
            coefficient: Coefficient of the original feed rate.
        """
        request = dto.TranslatorSetFeedRateOverrideRequest(
            translator_id=self.translator_id,
            coefficient=coefficient,
        )
        call_sync("gcode/set_feed_rate_override", request)

    @staticmethod
    def __free(
            translator_id: int
    ) -> None:
        """
        Releases native resources of a translator.

        Args:
            translator_id: The ID of the translator.
        """
        request = dto.TranslatorEmptyRequest(
            translator_id=translator_id,
        )
        call_sync("gcode/free", request)

    def __get_current_coordinate_system(
            self
    ) -> str:
        """
        Gets current coordinate system (e.g. G54).

        Returns:
            Current coordinate system.
        """
        request = dto.TranslatorEmptyRequest(
            translator_id=self.translator_id,
        )
        response = call_sync(
            "gcode/get_current_coordinate_system",
            request,
            dto.StringResponse.from_binary)
        return response.value

    def __del__(self) -> None:
        OfflineTranslator.__free(self.translator_id)



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict, Optional
import decimal
import zaber_bson
from ..axis_address import AxisAddress
from ..channel_address import ChannelAddress


@dataclass
class MicroscopeConfig:
    """
    Configuration representing a microscope setup.
    Device address of value 0 means that the part is not present.
    """

    focus_axis: Optional[AxisAddress] = None
    """
    Focus axis of the microscope.
    """

    x_axis: Optional[AxisAddress] = None
    """
    X axis of the microscope.
    """

    y_axis: Optional[AxisAddress] = None
    """
    Y axis of the microscope.
    """

    illuminator: Optional[int] = None
    """
    Illuminator device address.
    """

    filter_changer: Optional[int] = None
    """
    Filter changer device address.
    """

    objective_changer: Optional[int] = None
    """
    Objective changer device address.
    """

    autofocus: Optional[int] = None
    """
    Autofocus identifier.
    """

    camera_trigger: Optional[ChannelAddress] = None
    """
    Camera trigger digital output address.
    """

    @staticmethod
    def zero_values() -> 'MicroscopeConfig':
        return MicroscopeConfig(
            focus_axis=None,
            x_axis=None,
            y_axis=None,
            illuminator=None,
            filter_changer=None,
            objective_changer=None,
            autofocus=None,
            camera_trigger=None,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'MicroscopeConfig':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return MicroscopeConfig.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'focusAxis': self.focus_axis.to_dict() if self.focus_axis is not None else None,
            'xAxis': self.x_axis.to_dict() if self.x_axis is not None else None,
            'yAxis': self.y_axis.to_dict() if self.y_axis is not None else None,
            'illuminator': int(self.illuminator) if self.illuminator is not None else None,
            'filterChanger': int(self.filter_changer) if self.filter_changer is not None else None,
            'objectiveChanger': int(self.objective_changer) if self.objective_changer is not None else None,
            'autofocus': int(self.autofocus) if self.autofocus is not None else None,
            'cameraTrigger': self.camera_trigger.to_dict() if self.camera_trigger is not None else None,
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'MicroscopeConfig':
        return MicroscopeConfig(
            focus_axis=AxisAddress.from_dict(data.get('focusAxis')) if data.get('focusAxis') is not None else None,  # type: ignore
            x_axis=AxisAddress.from_dict(data.get('xAxis')) if data.get('xAxis') is not None else None,  # type: ignore
            y_axis=AxisAddress.from_dict(data.get('yAxis')) if data.get('yAxis') is not None else None,  # type: ignore
            illuminator=data.get('illuminator'),  # type: ignore
            filter_changer=data.get('filterChanger'),  # type: ignore
            objective_changer=data.get('objectiveChanger'),  # type: ignore
            autofocus=data.get('autofocus'),  # type: ignore
            camera_trigger=ChannelAddress.from_dict(data.get('cameraTrigger')) if data.get('cameraTrigger') is not None else None,  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.focus_axis is not None:
            if not isinstance(self.focus_axis, AxisAddress):
                raise ValueError(f'Property "FocusAxis" of "MicroscopeConfig" is not an instance of "AxisAddress".')

            self.focus_axis.validate()

        if self.x_axis is not None:
            if not isinstance(self.x_axis, AxisAddress):
                raise ValueError(f'Property "XAxis" of "MicroscopeConfig" is not an instance of "AxisAddress".')

            self.x_axis.validate()

        if self.y_axis is not None:
            if not isinstance(self.y_axis, AxisAddress):
                raise ValueError(f'Property "YAxis" of "MicroscopeConfig" is not an instance of "AxisAddress".')

            self.y_axis.validate()

        if self.illuminator is not None:
            if not isinstance(self.illuminator, (int, float, decimal.Decimal)):
                raise ValueError(f'Property "Illuminator" of "MicroscopeConfig" is not a number.')

            if int(self.illuminator) != self.illuminator:
                raise ValueError(f'Property "Illuminator" of "MicroscopeConfig" is not integer value.')

        if self.filter_changer is not None:
            if not isinstance(self.filter_changer, (int, float, decimal.Decimal)):
                raise ValueError(f'Property "FilterChanger" of "MicroscopeConfig" is not a number.')

            if int(self.filter_changer) != self.filter_changer:
                raise ValueError(f'Property "FilterChanger" of "MicroscopeConfig" is not integer value.')

        if self.objective_changer is not None:
            if not isinstance(self.objective_changer, (int, float, decimal.Decimal)):
                raise ValueError(f'Property "ObjectiveChanger" of "MicroscopeConfig" is not a number.')

            if int(self.objective_changer) != self.objective_changer:
                raise ValueError(f'Property "ObjectiveChanger" of "MicroscopeConfig" is not integer value.')

        if self.autofocus is not None:
            if not isinstance(self.autofocus, (int, float, decimal.Decimal)):
                raise ValueError(f'Property "Autofocus" of "MicroscopeConfig" is not a number.')

            if int(self.autofocus) != self.autofocus:
                raise ValueError(f'Property "Autofocus" of "MicroscopeConfig" is not integer value.')

        if self.camera_trigger is not None:
            if not isinstance(self.camera_trigger, ChannelAddress):
                raise ValueError(f'Property "CameraTrigger" of "MicroscopeConfig" is not an instance of "ChannelAddress".')

            self.camera_trigger.validate()



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional
import decimal
from collections.abc import Iterable
import zaber_bson
from .stream_segment_type import StreamSegmentType
from ..measurement import Measurement


@dataclass
class PvtPointRequest:

    interface_id: int = 0

    device: int = 0

    stream_id: int = 0

    pvt: bool = False

    type: StreamSegmentType = next(first for first in StreamSegmentType)

    positions: List[Measurement] = field(default_factory=list)

    velocities: List[Optional[Measurement]] = field(default_factory=list)

    time: Measurement = field(default_factory=Measurement.zero_values)

    @staticmethod
    def zero_values() -> 'PvtPointRequest':
        return PvtPointRequest(
            interface_id=0,
            device=0,
            stream_id=0,
            pvt=False,
            type=next(first for first in StreamSegmentType),
            positions=[],
            velocities=[],
            time=Measurement.zero_values(),
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'PvtPointRequest':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return PvtPointRequest.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'interfaceId': int(self.interface_id),
            'device': int(self.device),
            'streamId': int(self.stream_id),
            'pvt': bool(self.pvt),
            'type': self.type.value,
            'positions': [item.to_dict() for item in self.positions] if self.positions is not None else [],
            'velocities': [item.to_dict() if item is not None else None for item in self.velocities] if self.velocities is not None else [],
            'time': self.time.to_dict(),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'PvtPointRequest':
        return PvtPointRequest(
            interface_id=data.get('interfaceId'),  # type: ignore
            device=data.get('device'),  # type: ignore
            stream_id=data.get('streamId'),  # type: ignore
            pvt=data.get('pvt'),  # type: ignore
            type=StreamSegmentType(data.get('type')),  # type: ignore
            positions=[Measurement.from_dict(item) for item in data.get('positions')],  # type: ignore
            velocities=[Measurement.from_dict(item) if item is not None else None for item in data.get('velocities')],  # type: ignore
            time=Measurement.from_dict(data.get('time')),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.interface_id is None:
            raise ValueError(f'Property "InterfaceId" of "PvtPointRequest" is None.')

        if not isinstance(self.interface_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "InterfaceId" of "PvtPointRequest" is not a number.')

        if int(self.interface_id) != self.interface_id:
            raise ValueError(f'Property "InterfaceId" of "PvtPointRequest" is not integer value.')

        if self.device is None:
            raise ValueError(f'Property "Device" of "PvtPointRequest" is None.')

        if not isinstance(self.device, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "Device" of "PvtPointRequest" is not a number.')

        if int(self.device) != self.device:
            raise ValueError(f'Property "Device" of "PvtPointRequest" is not integer value.')

        if self.stream_id is None:
            raise ValueError(f'Property "StreamId" of "PvtPointRequest" is None.')

        if not isinstance(self.stream_id, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "StreamId" of "PvtPointRequest" is not a number.')

        if int(self.stream_id) != self.stream_id:
            raise ValueError(f'Property "StreamId" of "PvtPointRequest" is not integer value.')

        if self.type is None:
            raise ValueError(f'Property "Type" of "PvtPointRequest" is None.')

        if not isinstance(self.type, StreamSegmentType):
            raise ValueError(f'Property "Type" of "PvtPointRequest" is not an instance of "StreamSegmentType".')

        if self.positions is not None:
            if not isinstance(self.positions, Iterable):
                raise ValueError('Property "Positions" of "PvtPointRequest" is not iterable.')

            for i, positions_item in enumerate(self.positions):
                if positions_item is None:
                    raise ValueError(f'Item {i} in property "Positions" of "PvtPointRequest" is None.')

                if not isinstance(positions_item, Measurement):
                    raise ValueError(f'Item {i} in property "Positions" of "PvtPointRequest" is not an instance of "Measurement".')

                positions_item.validate()

        if self.velocities is not None:
            if not isinstance(self.velocities, Iterable):
                raise ValueError('Property "Velocities" of "PvtPointRequest" is not iterable.')

            for i, velocities_item in enumerate(self.velocities):
                if velocities_item is not None:
                    if not isinstance(velocities_item, Measurement):
                        raise ValueError(f'Item {i} in property "Velocities" of "PvtPointRequest" is not an instance of "Measurement".')

                    velocities_item.validate()

        if self.time is None:
            raise ValueError(f'Property "Time" of "PvtPointRequest" is None.')

        if not isinstance(self.time, Measurement):
            raise ValueError(f'Property "Time" of "PvtPointRequest" is not an instance of "Measurement".')

        self.time.validate()



# This file is generated. Do not modify by hand.
from enum import Enum


class DigitalOutputAction(Enum):
    """
    Action type for digital output.
    """

    OFF = 0
    ON = 1
    TOGGLE = 2
    KEEP = 3



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict, List
import decimal
from collections.abc import Iterable
import zaber_bson


@dataclass
class SetPeripheralStateExceptionData:
    """
    Contains additional data for a SetPeripheralStateFailedException.
    """

    axis_number: int
    """
    The number of axis where the exception originated.
    """

    settings: List[str]
    """
    A list of settings which could not be set.
    """

    servo_tuning: str
    """
    The reason servo tuning could not be set.
    """

    stored_positions: List[str]
    """
    The reasons stored positions could not be set.
    """

    storage: List[str]
    """
    The reasons storage could not be set.
    """

    @staticmethod
    def zero_values() -> 'SetPeripheralStateExceptionData':
        return SetPeripheralStateExceptionData(
            axis_number=0,
            settings=[],
            servo_tuning="",
            stored_positions=[],
            storage=[],
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'SetPeripheralStateExceptionData':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return SetPeripheralStateExceptionData.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'axisNumber': int(self.axis_number),
            'settings': [str(item or '') for item in self.settings] if self.settings is not None else [],
            'servoTuning': str(self.servo_tuning or ''),
            'storedPositions': [str(item or '') for item in self.stored_positions] if self.stored_positions is not None else [],
            'storage': [str(item or '') for item in self.storage] if self.storage is not None else [],
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'SetPeripheralStateExceptionData':
        return SetPeripheralStateExceptionData(
            axis_number=data.get('axisNumber'),  # type: ignore
            settings=data.get('settings'),  # type: ignore
            servo_tuning=data.get('servoTuning'),  # type: ignore
            stored_positions=data.get('storedPositions'),  # type: ignore
            storage=data.get('storage'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.axis_number is None:
            raise ValueError(f'Property "AxisNumber" of "SetPeripheralStateExceptionData" is None.')

        if not isinstance(self.axis_number, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "AxisNumber" of "SetPeripheralStateExceptionData" is not a number.')

        if int(self.axis_number) != self.axis_number:
            raise ValueError(f'Property "AxisNumber" of "SetPeripheralStateExceptionData" is not integer value.')

        if self.settings is not None:
            if not isinstance(self.settings, Iterable):
                raise ValueError('Property "Settings" of "SetPeripheralStateExceptionData" is not iterable.')

            for i, settings_item in enumerate(self.settings):
                if settings_item is not None:
                    if not isinstance(settings_item, str):
                        raise ValueError(f'Item {i} in property "Settings" of "SetPeripheralStateExceptionData" is not a string.')

        if self.servo_tuning is not None:
            if not isinstance(self.servo_tuning, str):
                raise ValueError(f'Property "ServoTuning" of "SetPeripheralStateExceptionData" is not a string.')

        if self.stored_positions is not None:
            if not isinstance(self.stored_positions, Iterable):
                raise ValueError('Property "StoredPositions" of "SetPeripheralStateExceptionData" is not iterable.')

            for i, stored_positions_item in enumerate(self.stored_positions):
                if stored_positions_item is not None:
                    if not isinstance(stored_positions_item, str):
                        raise ValueError(f'Item {i} in property "StoredPositions" of "SetPeripheralStateExceptionData" is not a string.')

        if self.storage is not None:
            if not isinstance(self.storage, Iterable):
                raise ValueError('Property "Storage" of "SetPeripheralStateExceptionData" is not iterable.')

            for i, storage_item in enumerate(self.storage):
                if storage_item is not None:
                    if not isinstance(storage_item, str):
                        raise ValueError(f'Item {i} in property "Storage" of "SetPeripheralStateExceptionData" is not a string.')



from ctypes import c_void_p, c_int, c_int64, c_uint8, CFUNCTYPE, cdll
from typing import Any
import os
import platform
import sys


def _load_library() -> Any:
    is64bit = sys.maxsize > 2**32
    os_system = platform.system().lower()
    os_machine = platform.machine().lower()

    if os_system == "darwin":
        arch = "uni"
    else:
        if os_machine.startswith("aarch64") or os_machine.startswith("arm"):
            arch = "arm64" if is64bit else "arm"
        else:
            arch = "amd64" if is64bit else "386"

    if os_system == "linux":
        ext = ".so"
    elif os_system == "darwin":
        ext = ".dylib"
    elif os_system == "windows":
        ext = ".dll"
    else:
        raise ImportError(f"Unsupported operating system {os_system}")

    lib_name = f"zaber-motion-core-{os_system}-{arch}{ext}"
    lib_path = os.path.join(os.path.dirname(__file__), "..", "zaber_motion_bindings", lib_name)
    lib_path = os.path.normpath(lib_path)

    if not os.path.exists(lib_path):
        lib_dir = os.path.dirname(lib_path)
        raise ImportError(f"Could not find library {lib_name} at path {lib_dir}. Please contact Zaber support.")
    return cdll.LoadLibrary(lib_path)


lib = _load_library()

CALLBACK = CFUNCTYPE(None, c_void_p, c_int64)

c_call = lib.zml_call
c_call.argtypes = [c_void_p, c_int64, CALLBACK, c_uint8]
c_call.restype = c_int

c_set_event_handler = lib.zml_setEventHandler
c_set_event_handler.argtypes = [c_int64, CALLBACK]
c_set_event_handler.restype = None



# This file is generated. Do not modify by hand.
# pylint: disable=line-too-long, unused-argument, f-string-without-interpolation, too-many-branches, too-many-statements, unnecessary-pass
from dataclasses import dataclass
from typing import Any, Dict
import decimal
import zaber_bson


@dataclass
class DeviceIOInfo:
    """
    Class representing information on the I/O channels of the device.
    """

    number_analog_outputs: int
    """
    Number of analog output channels.
    """

    number_analog_inputs: int
    """
    Number of analog input channels.
    """

    number_digital_outputs: int
    """
    Number of digital output channels.
    """

    number_digital_inputs: int
    """
    Number of digital input channels.
    """

    @staticmethod
    def zero_values() -> 'DeviceIOInfo':
        return DeviceIOInfo(
            number_analog_outputs=0,
            number_analog_inputs=0,
            number_digital_outputs=0,
            number_digital_inputs=0,
        )

    @staticmethod
    def from_binary(data_bytes: bytes) -> 'DeviceIOInfo':
        """" Deserialize a binary representation of this class. """
        data = zaber_bson.loads(data_bytes)  # type: Dict[str, Any]
        return DeviceIOInfo.from_dict(data)

    def to_binary(self) -> bytes:
        """" Serialize this class to a binary representation. """
        self.validate()
        return zaber_bson.dumps(self.to_dict())  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {
            'numberAnalogOutputs': int(self.number_analog_outputs),
            'numberAnalogInputs': int(self.number_analog_inputs),
            'numberDigitalOutputs': int(self.number_digital_outputs),
            'numberDigitalInputs': int(self.number_digital_inputs),
        }

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> 'DeviceIOInfo':
        return DeviceIOInfo(
            number_analog_outputs=data.get('numberAnalogOutputs'),  # type: ignore
            number_analog_inputs=data.get('numberAnalogInputs'),  # type: ignore
            number_digital_outputs=data.get('numberDigitalOutputs'),  # type: ignore
            number_digital_inputs=data.get('numberDigitalInputs'),  # type: ignore
        )

    def validate(self) -> None:
        """" Validates the properties of the instance. """
        if self.number_analog_outputs is None:
            raise ValueError(f'Property "NumberAnalogOutputs" of "DeviceIOInfo" is None.')

        if not isinstance(self.number_analog_outputs, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "NumberAnalogOutputs" of "DeviceIOInfo" is not a number.')

        if int(self.number_analog_outputs) != self.number_analog_outputs:
            raise ValueError(f'Property "NumberAnalogOutputs" of "DeviceIOInfo" is not integer value.')

        if self.number_analog_inputs is None:
            raise ValueError(f'Property "NumberAnalogInputs" of "DeviceIOInfo" is None.')

        if not isinstance(self.number_analog_inputs, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "NumberAnalogInputs" of "DeviceIOInfo" is not a number.')

        if int(self.number_analog_inputs) != self.number_analog_inputs:
            raise ValueError(f'Property "NumberAnalogInputs" of "DeviceIOInfo" is not integer value.')

        if self.number_digital_outputs is None:
            raise ValueError(f'Property "NumberDigitalOutputs" of "DeviceIOInfo" is None.')

        if not isinstance(self.number_digital_outputs, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "NumberDigitalOutputs" of "DeviceIOInfo" is not a number.')

        if int(self.number_digital_outputs) != self.number_digital_outputs:
            raise ValueError(f'Property "NumberDigitalOutputs" of "DeviceIOInfo" is not integer value.')

        if self.number_digital_inputs is None:
            raise ValueError(f'Property "NumberDigitalInputs" of "DeviceIOInfo" is None.')

        if not isinstance(self.number_digital_inputs, (int, float, decimal.Decimal)):
            raise ValueError(f'Property "NumberDigitalInputs" of "DeviceIOInfo" is not a number.')

        if int(self.number_digital_inputs) != self.number_digital_inputs:
            raise ValueError(f'Property "NumberDigitalInputs" of "DeviceIOInfo" is not integer value.')
